# creates: test.txt
import runpy
import numpy as np
from pathlib import Path

# Monkey-patch view() to avoid ASE-GUI windows popping up:
import ase.visualize
ase.visualize.view = lambda *args, **kwargs: None


def run(script):
    return runpy.run_path(Path('selfdiffusion') / script)


def run_and_get_ptp_energy(script):
    dct = run(script)
    energy = np.ptp([i.get_potential_energy() for i in dct['images']])
    return energy


e1 = run_and_get_ptp_energy('neb1.py')
assert abs(e1 - 0.111) < 0.002
e2 = run_and_get_ptp_energy('neb2.py')
assert abs(e2 - 0.564) < 0.002
e3 = run_and_get_ptp_energy('neb3.py')
assert abs(e3 - 0.239) < 0.002

with open('test.txt', 'w') as fd:
    print(e1, e2, e3, file=fd)



# creates: C2H4.png
from ase.build.molecule import molecule
from ase.io import write
from ase.io.pov import get_bondpairs, set_high_bondorder_pairs
C2H4 = molecule('C2H4')
r = [{'C': 0.4, 'H': 0.2}[at.symbol] for at in C2H4]
bondpairs = get_bondpairs(C2H4, radius=1.1)
high_bondorder_pairs = {}
# This defines offset, bond order, and bond_offset of the bond between 0 and 1
high_bondorder_pairs[(0, 1)] = ((0, 0, 0), 2, (0.17, 0.17, 0))
bondpairs = set_high_bondorder_pairs(bondpairs, high_bondorder_pairs)

renderer = write('C2H4.pov', C2H4, format='pov',
                 radii=r, rotation='90y',
                 povray_settings=dict(canvas_width=200, bondatoms=bondpairs))

renderer.render()



import pytest
import numpy as np

from ase import Atoms
from ase.io import read, write


@pytest.mark.parametrize('pbc', [False, [True, True, False], True])
@pytest.mark.parametrize('cell', [None, [[2.5, 0, 0], [2, 4, 0], [1, 2, 3]]])
@pytest.mark.parametrize('write_format', ['gen', 'dftb'])
def test_gen(pbc, cell, write_format):
    atoms = Atoms(symbols='OCO', pbc=pbc, cell=cell,
                  positions=[[-0.1, 1.2, 0.3],
                             [-0.1, 0.0, 0.2],
                             [0.4, -0.9, 0.0]])
    write('test.gen', atoms, format=write_format)

    atoms_new = read('test.gen')
    assert np.all(atoms_new.numbers == atoms.numbers)
    assert np.allclose(atoms_new.positions, atoms.positions)

    if atoms.pbc.any():
        assert np.all(atoms_new.pbc)
        if atoms.cell is not None:
            assert np.allclose(atoms_new.cell, atoms.cell)
    else:
        assert np.all(~atoms_new.pbc)
        assert np.allclose(atoms_new.cell, 0.)


def test_gen_multiple():
    # Try with multiple images. This is not supported by the
    # format and should fail
    atoms = Atoms('H2')

    with pytest.raises(ValueError):
        write('test.gen', [atoms, atoms])



import os
from pathlib import Path
import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal

from ase import units, Atoms
import ase.io
from ase.calculators.qmmm import ForceConstantCalculator
from ase.vibrations import Vibrations, VibrationsData
from ase.thermochemistry import IdealGasThermo


class TestHarmonicVibrations:
    """Test the ase.vibrations.Vibrations object using a harmonic calculator
    """
    def setup(self):
        self.logfile = 'vibrations-log.txt'

    @pytest.fixture
    def random_dimer(self):
        rng = np.random.RandomState(42)

        d = 1 + 0.5 * rng.rand()
        z_values = rng.randint(1, high=50, size=2)

        hessian = rng.rand(6, 6)
        hessian += hessian.T  # Ensure the random Hessian is symmetric

        atoms = Atoms(z_values, [[0, 0, 0], [0, 0, d]])
        ref_atoms = atoms.copy()
        atoms.calc = ForceConstantCalculator(D=hessian,
                                             ref=ref_atoms,
                                             f0=np.zeros((2, 3)))
        return atoms

    def test_harmonic_vibrations(self, testdir):
        """Check the numerics with a trivial case: one atom in harmonic well"""
        rng = np.random.RandomState(42)

        k = rng.rand()

        ref_atoms = Atoms('H', positions=np.zeros([1, 3]))
        atoms = ref_atoms.copy()
        mass = atoms.get_masses()[0]

        atoms.calc = ForceConstantCalculator(D=np.eye(3) * k,
                                             ref=ref_atoms,
                                             f0=np.zeros((1, 3)))
        vib = Vibrations(atoms, name='harmonic')
        vib.run()
        vib.read()

        expected_energy = (units._hbar  # In J/s
                           * np.sqrt(k  # In eV/A^2
                                     * units._e  # eV -> J
                                     * units.m**2  # A^-2 -> m^-2
                                     / mass  # in amu
                                     / units._amu  # amu^-1 -> kg^-1
                                     )
                           ) / units._e  # J/s -> eV/s

        assert np.allclose(vib.get_energies(), expected_energy)

    def test_consistency_with_vibrationsdata(self, testdir, random_dimer):
        vib = Vibrations(random_dimer, delta=1e-6, nfree=4)
        vib.run()
        vib_data = vib.get_vibrations()

        assert_array_almost_equal(vib.get_energies(),
                                  vib_data.get_energies())

        for mode_index in range(3 * len(vib.atoms)):
            assert_array_almost_equal(vib.get_mode(mode_index),
                                      vib_data.get_modes()[mode_index])

        # Hessian should be close to the ForceConstantCalculator input
        assert_array_almost_equal(random_dimer.calc.D,
                                  vib_data.get_hessian_2d(),
                                  decimal=6)

    def test_json_manipulation(self, testdir, random_dimer):
        vib = Vibrations(random_dimer, name='interrupt')
        vib.run()

        disp_file = Path('interrupt/cache.1x-.json')
        comb_file = Path('interrupt/combined.json')
        assert disp_file.is_file()
        assert not comb_file.is_file()

        # Should do nothing harmful as files are already split
        # (It used to raise an error but this is no longer implemented.)
        vib.split()

        # Build a combined file
        assert vib.combine() == 13

        # Individual displacements should be gone, combination should exist
        assert not disp_file.is_file()
        assert comb_file.is_file()

        # Not allowed to run after data has been combined
        with pytest.raises(RuntimeError):
            vib.run()
        # But reading is allowed
        vib.read()

        # Splitting should fail if any split file already exists
        with open(disp_file, 'w') as fd:
            fd.write("hello")

        with pytest.raises(AssertionError):
            vib.split()

        os.remove(disp_file)

        # Now split() for real: replace .all.json file with displacements
        vib.split()
        assert disp_file.is_file()
        assert not comb_file.is_file()

    def test_vibrations_methods(self, testdir, random_dimer):
        vib = Vibrations(random_dimer)
        vib.run()
        vib_energies = vib.get_energies()

        for image in vib.iterimages():
            assert len(image) == 2

        thermo = IdealGasThermo(vib_energies=vib_energies, geometry='linear',
                                atoms=vib.atoms, symmetrynumber=2, spin=0)
        thermo.get_gibbs_energy(temperature=298.15, pressure=2 * 101325.,
                                verbose=False)

        with open(self.logfile, 'w') as fd:
            vib.summary(log=fd)

        with open(self.logfile, 'rt') as fd:
            log_txt = fd.read()
            assert log_txt == '\n'.join(
                VibrationsData._tabulate_from_energies(vib_energies)) + '\n'

        last_mode = vib.get_mode(-1)
        scale = 0.5
        assert_array_almost_equal(vib.show_as_force(-1, scale=scale,
                                                    show=False).get_forces(),
                                  last_mode * 3 * len(vib.atoms) * scale)

        vib.write_mode(n=3, nimages=5)
        for i in range(3):
            assert not Path('vib.{}.traj'.format(i)).is_file()
        mode_traj = ase.io.read('vib.3.traj', index=':')
        assert len(mode_traj) == 5

        assert_array_almost_equal(mode_traj[0].get_all_distances(),
                                  random_dimer.get_all_distances())
        with pytest.raises(AssertionError):
            assert_array_almost_equal(mode_traj[4].get_all_distances(),
                                      random_dimer.get_all_distances())

        assert vib.clean(empty_files=True) == 0
        assert vib.clean() == 13
        assert len(list(vib.iterimages())) == 13

        d = dict(vib.iterdisplace(inplace=False))

        for name, image in vib.iterdisplace(inplace=True):
            assert d[name] == random_dimer

    def test_vibrations_restart_dir(self, testdir, random_dimer):
        vib = Vibrations(random_dimer)
        vib.run()
        freqs = vib.get_frequencies()
        assert freqs is not None

        # write/read the data from another working directory
        atoms = random_dimer.copy()  # This copy() removes the Calculator

        with ase.utils.workdir('run_from_here', mkdir=True):
            vib = Vibrations(atoms, name=str(Path.cwd().parent / 'vib'))
            assert_array_almost_equal(freqs, vib.get_frequencies())
            assert vib.clean() == 13


class TestVibrationsDataStaticMethods:
    @pytest.mark.parametrize('mask,expected_indices',
                             [([True, True, False, True], [0, 1, 3]),
                              ([False, False], []),
                              ([], []),
                              (np.array([True, True]), [0, 1]),
                              (np.array([False, True, True]), [1, 2]),
                              (np.array([], dtype=bool), [])])
    def test_indices_from_mask(self, mask, expected_indices):
        assert VibrationsData.indices_from_mask(mask) == expected_indices

    def test_tabulate_energies(self):
        # Test the private classmethod _tabulate_from_energies
        # used by public tabulate() method
        energies = np.array([1., complex(2., 1.), complex(1., 1e-3)])

        table = VibrationsData._tabulate_from_energies(energies, im_tol=1e-2)

        for sep_row in 0, 2, 6:
            assert table[sep_row] == '-' * 21
        assert tuple(table[1].strip().split()) == ('#', 'meV', 'cm^-1')

        expected_rows = [
            # energy in eV should be converted to meV and cm-1
            ('0', '1000.0', '8065.5'),
            # Imaginary component over threshold detected
            ('1', '1000.0i', '8065.5i'),
            # Small imaginary component ignored
            ('2', '1000.0', '8065.5')]

        for row, expected in zip(table[3:6], expected_rows):
            assert tuple(row.split()) == expected

        # ZPE = (1 + 2 + 1) / 2  - currently we keep all real parts
        assert table[7].split()[2] == '2.000'
        assert len(table) == 8

    na2 = Atoms('Na2', cell=[2, 2, 2], positions=[[0, 0, 0],
                                                  [1, 1, 1]])
    na2_image_1 = na2.copy()
    na2_image_1.info.update({'mode#': '0',
                             'frequency_cm-1': 8065.5})
    na2_image_1.arrays['mode'] = np.array([[1., 1., 1.],
                                           [0.5, 0.5, 0.5]])

    @pytest.mark.parametrize('kwargs,expected',
                             [(dict(atoms=na2,
                                    energies=[1.],
                                    modes=np.array([[[1., 1., 1.],
                                                     [0.5, 0.5, 0.5]]])),
                               [na2_image_1])
                              ])
    def test_get_jmol_images(self, kwargs, expected):
        # Test the private staticmethod _get_jmol_images
        # used by the public write_jmol_images() method
        from ase.calculators.calculator import compare_atoms

        jmol_images = list(VibrationsData._get_jmol_images(**kwargs))
        assert len(jmol_images) == len(expected)

        for image, reference in zip(jmol_images, expected):
            assert compare_atoms(image, reference) == []
            for key, value in reference.info.items():
                if key == 'frequency_cm-1':
                    assert float(image.info[key]) == pytest.approx(value,
                                                                   abs=0.1)
                else:
                    assert image.info[key] == value


class TestVibrationsData:
    @pytest.fixture
    def random_dimer(self):
        rng = np.random.RandomState(42)

        d = 1 + 0.5 * rng.rand()
        z_values = rng.randint(1, high=50, size=2)

        hessian = rng.rand(6, 6)
        hessian += hessian.T  # Ensure the random Hessian is symmetric

        atoms = Atoms(z_values, [[0, 0, 0], [0, 0, d]])
        ref_atoms = atoms.copy()
        atoms.calc = ForceConstantCalculator(D=hessian,
                                             ref=ref_atoms,
                                             f0=np.zeros((2, 3)))
        return atoms

    @pytest.fixture
    def n2_data(self):
        return{'atoms': Atoms('N2', positions=[[0., 0., 0.05095057],
                                               [0., 0., 1.04904943]]),
               'hessian': np.array([[[[4.67554672e-03, 0.0, 0.0],
                                      [-4.67554672e-03, 0.0, 0.0]],

                                     [[0.0, 4.67554672e-03, 0.0],
                                      [0.0, -4.67554672e-03, 0.0]],

                                     [[0.0, 0.0, 3.90392599e+01],
                                      [0.0, 0.0, -3.90392599e+01]]],

                                    [[[-4.67554672e-03, 0.0, 0.0],
                                      [4.67554672e-03, 0.0, 0.0]],

                                     [[0.0, -4.67554672e-03, 0.0],
                                      [0.0, 4.67554672e-03, 0.0]],

                                     [[0.0, 0.0, -3.90392599e+01],
                                      [0.0, 0.0, 3.90392599e+01]]]]),
               'ref_frequencies': [0.00000000e+00 + 0.j,
                                   6.06775530e-08 + 0.j,
                                   3.62010442e-06 + 0.j,
                                   1.34737571e+01 + 0.j,
                                   1.34737571e+01 + 0.j,
                                   1.23118496e+03 + 0.j],
               'ref_zpe': 0.07799427233401508,
               'ref_forces': np.array([[0., 0., -2.26722e-1],
                                       [0., 0., 2.26722e-1]])
               }

    @pytest.fixture
    def n2_unstable_data(self):
        return{'atoms': Atoms('N2', positions=[[0., 0., 0.45],
                                               [0., 0., -0.45]]),
               'hessian': np.array(
                   [-5.150829928323684, 0.0, -0.6867385017096544,
                    5.150829928323684, 0.0, 0.6867385017096544, 0.0,
                    -5.158454318599951, 0.0, 0.0, 5.158454318599951, 0.0,
                    -0.6867385017096544, 0.0, 56.65107699250456,
                    0.6867385017096544, 0.0, -56.65107699250456,
                    5.150829928323684, 0.0, 0.6867385017096544,
                    -5.150829928323684, 0.0, -0.6867385017096544, 0.0,
                    5.158454318599951, 0.0, 0.0, -5.158454318599951, 0.0,
                    0.6867385017096544, 0.0, -56.65107699250456,
                    -0.6867385017096544, 0.0, 56.65107699250456
                    ]).reshape((2, 3, 2, 3))
               }

    @pytest.fixture
    def n2_vibdata(self, n2_data):
        return VibrationsData(n2_data['atoms'], n2_data['hessian'])

    def setup(self):
        self.jmol_file = 'vib-data.xyz'

    def test_init(self, n2_data):
        # Check that init runs without error; properties are checked in other
        # methods using the (identical) n2_vibdata fixture
        VibrationsData(n2_data['atoms'], n2_data['hessian'])

    def test_energies_and_modes(self, n2_data, n2_vibdata):
        energies, modes = n2_vibdata.get_energies_and_modes()
        assert_array_almost_equal(n2_data['ref_frequencies'],
                                  energies / units.invcm,
                                  decimal=5)
        assert_array_almost_equal(n2_data['ref_frequencies'],
                                  n2_vibdata.get_energies() / units.invcm,
                                  decimal=5)
        assert_array_almost_equal(n2_data['ref_frequencies'],
                                  n2_vibdata.get_frequencies(),
                                  decimal=5)

        assert (n2_vibdata.get_zero_point_energy()
                == pytest.approx(n2_data['ref_zpe']))

        assert n2_vibdata.tabulate() == (
            '\n'.join(VibrationsData._tabulate_from_energies(energies)) + '\n')

        atoms_with_forces = n2_vibdata.show_as_force(-1, show=False)

        try:
            assert_array_almost_equal(atoms_with_forces.get_forces(),
                                      n2_data['ref_forces'])
        except AssertionError:
            # Eigenvectors may be off by a sign change, which is allowed
            assert_array_almost_equal(atoms_with_forces.get_forces(),
                                      -n2_data['ref_forces'])

    def test_imaginary_energies(self, n2_unstable_data):
        vib_data = VibrationsData(n2_unstable_data['atoms'],
                                  n2_unstable_data['hessian'])

        assert vib_data.tabulate() == (
            '\n'.join(VibrationsData._tabulate_from_energies(
                vib_data.get_energies()))
            + '\n')

    def test_zero_mass(self, n2_data):
        atoms = n2_data['atoms']
        atoms.set_masses([0., 1.])
        vib_data = VibrationsData(atoms, n2_data['hessian'])
        with pytest.raises(ValueError):
            vib_data.get_energies_and_modes()

    def test_new_mass(self, n2_data, n2_vibdata):
        original_masses = n2_vibdata.get_atoms().get_masses()
        new_masses = original_masses * 3
        new_vib_data = n2_vibdata.with_new_masses(new_masses)
        assert_array_almost_equal(new_vib_data.get_atoms().get_masses(),
                                  new_masses)
        assert_array_almost_equal(n2_vibdata.get_energies() / np.sqrt(3),
                                  new_vib_data.get_energies())

    def test_fixed_atoms(self, n2_data):
        vib_data = VibrationsData(n2_data['atoms'],
                                  n2_data['hessian'][1:, :, 1:, :],
                                  indices=[1, ])
        assert vib_data.get_indices() == [1, ]
        assert vib_data.get_mask().tolist() == [False, True]

    def test_dos(self, n2_vibdata):
        with pytest.warns(np.ComplexWarning):
            dos = n2_vibdata.get_dos()
        assert_array_almost_equal(dos.get_energies(),
                                  n2_vibdata.get_energies())

    def test_pdos(self, n2_vibdata):
        with pytest.warns(np.ComplexWarning):
            pdos = n2_vibdata.get_pdos()
        assert_array_almost_equal(pdos[0].get_energies(),
                                  n2_vibdata.get_energies())
        assert_array_almost_equal(pdos[1].get_energies(),
                                  n2_vibdata.get_energies())
        # 3N states = 6, divided equally over two N atoms = 3.0
        assert sum(pdos[0].get_weights()) == pytest.approx(3.0)

    def test_todict(self, n2_data, n2_vibdata):
        vib_data_dict = n2_vibdata.todict()

        assert vib_data_dict['indices'] is None
        assert_array_almost_equal(vib_data_dict['atoms'].positions,
                                  n2_data['atoms'].positions)
        assert_array_almost_equal(vib_data_dict['hessian'],
                                  n2_data['hessian'])

    def test_dict_roundtrip(self, n2_vibdata):
        vib_data_dict = n2_vibdata.todict()
        vib_data_roundtrip = VibrationsData.fromdict(vib_data_dict)

        for getter in ('get_atoms',):
            assert (getattr(n2_vibdata, getter)()
                    == getattr(vib_data_roundtrip, getter)())
        for array_getter in ('get_hessian', 'get_hessian_2d',
                             'get_mask', 'get_indices'):
            assert_array_almost_equal(
                getattr(n2_vibdata, array_getter)(),
                getattr(vib_data_roundtrip, array_getter)())

    @pytest.mark.parametrize('indices, expected_mask',
                             [([1], [False, True]),
                              (None, [True, True])])
    def test_dict_indices(self, n2_vibdata, indices, expected_mask):
        vib_data_dict = n2_vibdata.todict()
        vib_data_dict['indices'] = indices

        # Reduce size of Hessian if necessary
        if indices is not None:
            n_active = len(indices)
            vib_data_dict['hessian'] = (
                np.asarray(vib_data_dict['hessian']
                           )[:n_active, :, :n_active, :].tolist())

        vib_data_fromdict = VibrationsData.fromdict(vib_data_dict)
        assert_array_almost_equal(vib_data_fromdict.get_mask(), expected_mask)

    def test_jmol_roundtrip(self, testdir, n2_data):
        ir_intensities = np.random.RandomState(42).rand(6)

        vib_data = VibrationsData(n2_data['atoms'], n2_data['hessian'])
        vib_data.write_jmol(self.jmol_file, ir_intensities=ir_intensities)

        images = ase.io.read(self.jmol_file, index=':')
        for i, image in enumerate(images):
            assert_array_almost_equal(image.positions,
                                      vib_data.get_atoms().positions)
            assert (image.info['IR_intensity']
                    == pytest.approx(ir_intensities[i]))
            assert_array_almost_equal(image.arrays['mode'],
                                      vib_data.get_modes()[i])

    def test_bad_hessian(self, n2_data):
        bad_hessians = (None, 'fish', 1,
                        np.array([1, 2, 3]),
                        np.eye(6),
                        np.array([[[1, 0, 0]],
                                  [[0, 0, 1]]]))

        for bad_hessian in bad_hessians:
            with pytest.raises(ValueError):
                VibrationsData(n2_data['atoms'], bad_hessian)

    def test_bad_hessian2d(self, n2_data):
        bad_hessians = (None, 'fish', 1,
                        np.array([1, 2, 3]),
                        n2_data['hessian'],
                        np.array([[[1, 0, 0]],
                                  [[0, 0, 1]]]))

        for bad_hessian in bad_hessians:
            with pytest.raises(ValueError):
                VibrationsData.from_2d(n2_data['atoms'], bad_hessian)


class TestSlab:
    "N2 above Ag slab - vibration with frozen molecules"
    def test_vibration_on_surface(self, testdir):
        from ase.build import fcc111, add_adsorbate
        ag_slab = fcc111('Ag', (4, 4, 2), a=2)
        n2 = Atoms('N2', positions=[[0., 0., 0.],
                                    [0., np.sqrt(2), np.sqrt(2)]])
        add_adsorbate(ag_slab, n2, height=1, position='fcc')

        # Add an interaction between the N atoms
        hessian_bottom_corner = np.zeros((2, 3, 2, 3))
        hessian_bottom_corner[-1, :, -2] = [1, 1, 1]
        hessian_bottom_corner[-2, :, -1] = [1, 1, 1]

        hessian = np.zeros((34, 3, 34, 3))
        hessian[32:, :, 32:, :] = hessian_bottom_corner

        ag_slab.calc = ForceConstantCalculator(hessian.reshape((34 * 3,
                                                                34 * 3)),
                                               ref=ag_slab.copy(),
                                               f0=np.zeros((34, 3)))

        # Check that Vibrations with restricted indices returns correct Hessian
        vibs = Vibrations(ag_slab, indices=[-2, -1])
        vibs.run()
        vibs.read()

        assert_array_almost_equal(vibs.get_vibrations().get_hessian(),
                                  hessian_bottom_corner)

        # These should blow up if the vectors don't match number of atoms
        vibs.summary()
        vibs.write_jmol()

        for i in range(6):
            # Frozen atoms should have zero displacement
            assert_array_almost_equal(vibs.get_mode(i)[0], [0., 0., 0.])

            # The N atoms should have finite displacement
            assert np.all(vibs.get_mode(i)[-2:, :])



from .vasp import Vasp
from .vasp_auxiliary import (get_vasp_version, VaspChargeDensity,
                             VaspDos, xdat2traj)
from .vasp2 import Vasp2
from .interactive import VaspInteractive
__all__ = [
    'Vasp', 'get_vasp_version', 'VaspChargeDensity', 'VaspDos', 'xdat2traj',
    'VaspInteractive', 'Vasp2',
]



import numpy as np

from ase.calculators.siesta.parameters import Species, PAOBasisBlock
from ase import Atoms


def test_siesta(siesta_factory):
    # Setup test structures.
    h = Atoms('H', [(0.0, 0.0, 0.0)])
    ch4 = Atoms('CH4', np.array([
        [0.000000, 0.000000, 0.000000],
        [0.682793, 0.682793, 0.682793],
        [-0.682793, -0.682793, 0.682790],
        [-0.682793, 0.682793, -0.682793],
        [0.682793, -0.682793, -0.682793]]))

    siesta = siesta_factory.calc()

    # Test simple fdf-argument case.
    atoms = h.copy()
    siesta = siesta_factory.calc(
        label='test_label',
        fdf_arguments={'DM.Tolerance': 1e-3})
    atoms.calc = siesta
    siesta.write_input(atoms, properties=['energy'])
    atoms = h.copy()
    atoms.calc = siesta
    siesta.write_input(atoms, properties=['energy'])
    with open('test_label.fdf', 'r') as fd:
        lines = fd.readlines()
    assert any([line.split() == ['DM.Tolerance', '0.001'] for line in lines])

    # Test (slightly) more complex case of setting fdf-arguments.
    siesta = siesta_factory.calc(
        label='test_label',
        mesh_cutoff=3000,
        fdf_arguments={
            'DM.Tolerance': 1e-3,
            'ON.eta': (5, 'Ry')})
    atoms.calc = siesta
    siesta.write_input(atoms, properties=['energy'])
    atoms = h.copy()
    atoms.calc = siesta
    siesta.write_input(atoms, properties=['energy'])
    with open('test_label.fdf', 'r') as fd:
        lines = fd.readlines()

    assert 'MeshCutoff\t3000\teV\n' in lines
    assert 'DM.Tolerance\t0.001\n' in lines
    assert 'ON.eta\t5\tRy\n' in lines

    # Test setting fdf-arguments after initiation.
    siesta.set_fdf_arguments(
        {'DM.Tolerance': 1e-2,
         'ON.eta': (2, 'Ry')})
    siesta.write_input(atoms, properties=['energy'])
    with open('test_label.fdf', 'r') as fd:
        lines = fd.readlines()
    assert 'MeshCutoff\t3000\teV\n' in lines
    assert 'DM.Tolerance\t0.01\n' in lines
    assert 'ON.eta\t2\tRy\n' in lines

    # Test initiation using Species.
    atoms = ch4.copy()
    species, numbers = siesta.species(atoms)
    assert all(numbers == np.array([1, 2, 2, 2, 2]))
    siesta = siesta_factory.calc(species=[Species(symbol='C', tag=1)])
    species, numbers = siesta.species(atoms)
    assert all(numbers == np.array([1, 2, 2, 2, 2]))
    atoms.set_tags([0, 0, 0, 1, 0])
    species, numbers = siesta.species(atoms)
    assert all(numbers == np.array([1, 2, 2, 2, 2]))
    siesta = siesta_factory.calc(species=[Species(symbol='H', tag=1, basis_set='SZ')])
    species, numbers = siesta.species(atoms)
    assert all(numbers == np.array([1, 2, 2, 3, 2]))
    siesta = siesta_factory.calc(label='test_label', species=species)
    siesta.write_input(atoms, properties=['energy'])
    with open('test_label.fdf', 'r') as fd:
        lines = fd.readlines()
    lines = [line.split() for line in lines]
    assert ['1', '6', 'C.lda.1'] in lines
    assert ['2', '1', 'H.lda.2'] in lines
    assert ['3', '1', 'H.lda.3'] in lines
    assert ['C.lda.1', 'DZP'] in lines
    assert ['H.lda.2', 'DZP'] in lines
    assert ['H.lda.3', 'SZ'] in lines

    # Test if PAO block can be given as species.
    c_basis = """2 nodes 1.00
    0 1 S 0.20 P 1 0.20 6.00
    5.00
    1.00
    1 2 S 0.20 P 1 E 0.20 6.00
    6.00 5.00
    1.00 0.95"""
    basis_set = PAOBasisBlock(c_basis)
    species = Species(symbol='C', basis_set=basis_set)
    siesta = siesta_factory.calc(label='test_label', species=[species])
    siesta.write_input(atoms, properties=['energy'])
    with open('test_label.fdf', 'r') as fd:
        lines = fd.readlines()
    lines = [line.split() for line in lines]
    assert ['%block', 'PAO.Basis'] in lines
    assert ['%endblock', 'PAO.Basis'] in lines



from io import StringIO
from ase.io import read
from ase.utils import reader

# Made from NWChem interface


@reader
def read_geom_orcainp(fd):
    """Method to read geometry from an ORCA input file."""
    lines = fd.readlines()

    # Find geometry region of input file.
    stopline = 0
    for index, line in enumerate(lines):
        if line[1:].startswith('xyz '):
            startline = index + 1
            stopline = -1
        elif (line.startswith('end') and stopline == -1):
            stopline = index
        elif (line.startswith('*') and stopline == -1):
            stopline = index
    # Format and send to read_xyz.
    xyz_text = '%i\n' % (stopline - startline)
    xyz_text += ' geometry\n'
    for line in lines[startline:stopline]:
        xyz_text += line
    atoms = read(StringIO(xyz_text), format='xyz')
    atoms.set_cell((0., 0., 0.))  # no unit cell defined

    return atoms


def write_orca(atoms, **params):
    """Function to write ORCA input file
    """
    charge = params['charge']
    mult = params['mult']
    label = params['label']

    if 'pcpot' in params.keys():
        pcpot = params['pcpot']
        pcstring = '% pointcharges \"' +\
                   label + '.pc\"\n\n'
        params['orcablocks'] += pcstring
        pcpot.write_mmcharges(label)

    with open(label + '.inp', 'w') as fd:
        fd.write("! engrad %s \n" % params['orcasimpleinput'])
        fd.write("%s \n" % params['orcablocks'])

        fd.write('*xyz')
        fd.write(" %d" % charge)
        fd.write(" %d \n" % mult)
        for atom in atoms:
            if atom.tag == 71:  # 71 is ascii G (Ghost)
                symbol = atom.symbol + ' : '
            else:
                symbol = atom.symbol + '   '
            fd.write(symbol +
                     str(atom.position[0]) + ' ' +
                     str(atom.position[1]) + ' ' +
                     str(atom.position[2]) + '\n')
        fd.write('*\n')



import pytest
import numpy as np

from ase.outputs import Properties, all_outputs


@pytest.fixture
def rng():
    return np.random.RandomState(17)


def test_properties_big(rng):
    nspins, nkpts, nbands = 2, 3, 5
    natoms = 4

    results = dict(
        natoms=natoms,
        energy=rng.rand(),
        free_energy=rng.rand(),
        energies=rng.rand(natoms),
        forces=rng.rand(natoms, 3),
        stress=rng.rand(6),
        stresses=rng.rand(natoms, 6),
        nspins=nspins,
        nkpts=nkpts,
        nbands=nbands,
        eigenvalues=rng.rand(nspins, nkpts, nbands),
        occupations=rng.rand(nspins, nkpts, nbands),
        fermi_level=rng.rand(),
        ibz_kpoints=rng.rand(nkpts, 3),
        kpoint_weights=rng.rand(nkpts),
    )

    props = Properties(results)
    #assert set(out) == all_outputs, all_outputs ^ set(out)

    for name in all_outputs:
        assert name in props, name
        obj = props[name]
        #obj = getattr(out, name)
        print(name, obj)



def test_cutoffs():
    from ase.neighborlist import natural_cutoffs
    from ase import Atoms
    import numpy as np

    atoms = Atoms("HCOPtAu")

    assert np.allclose(natural_cutoffs(atoms), [0.31, 0.76, 0.66, 1.36, 1.36])
    assert np.allclose(natural_cutoffs(atoms, mult=1.2), [0.372, 0.912, 0.792,
                                                          1.632, 1.632])
    assert np.allclose(natural_cutoffs(atoms, mult=1.2, Au=1), [0.372, 0.912,
                                                                0.792, 1.632, 1])



from ase.build import bulk
from ase.optimize import BFGS
from ase.calculators.abinit import Abinit
from ase.calculators.socketio import SocketIOCalculator
from ase.constraints import ExpCellFilter


atoms = bulk('Si')
atoms.rattle(stdev=0.1, seed=42)

# Configuration parameters; please edit as appropriate
pps = '/path/to/pseudopotentials'
pseudopotentials = {'Si': '14-Si.LDA.fhi'}
exe = 'abinit'

unixsocket = 'ase_abinit'
command = f'{exe} PREFIX.in --ipi {unixsocket}:UNIX > PREFIX.log'
# (In the command, note that PREFIX.in must precede --ipi.)


configuration_kwargs = dict(
    command=command,
    pp_paths=[pps],
    v8_legacy_format=False,
)


# Implementation note: Socket-driven calculations in Abinit inherit several
# controls for from the ordinary cell optimization code.  We have to hack those
# variables in order for Abinit not to decide that the calculation converged:
boilerplate_kwargs = dict(
    ionmov=28,  # activate i-pi/socket mode
    expert_user=1,  # Ignore warnings (chksymbreak, chksymtnons, chkdilatmx)
    optcell=2,  # allow the cell to relax
    tolmxf=1e-300,  # Prevent Abinit from thinking we "converged"
    ntime=100_000,  # Allow "infinitely" many iterations in Abinit
    ecutsm=0.5,  # Smoothing PW cutoff energy (mandatory for cell optimization)
)


kwargs = dict(
    ecut=5 * 27.3,
    tolvrs=1e-8,
    kpts=[2, 2, 2],
    **boilerplate_kwargs,
    **configuration_kwargs,
)

abinit = Abinit(**kwargs)

opt = BFGS(ExpCellFilter(atoms),
           trajectory='opt.traj')

with SocketIOCalculator(abinit, unixsocket=unixsocket) as atoms.calc:
    opt.run(fmax=0.01)



import importlib
from typing import List, Tuple
from ase.utils import search_current_git_hash


def format_dependency(modname: str) -> Tuple[str, str]:
    """Return (name, path) for given module."""
    try:
        module = importlib.import_module(modname)
    except ImportError:
        return modname, 'not installed'

    version = getattr(module, '__version__', '?')
    name = f'{modname}-{version}'
    if modname == 'ase':
        githash = search_current_git_hash(module)
        if githash:
            name += '-{:.10}'.format(githash)

    # (only packages have __path__, but we are importing packages.)
    return name, str(module.__path__[0])  # type: ignore


def all_dependencies() -> List[Tuple[str, str]]:
    names = ['ase', 'numpy', 'scipy', 'matplotlib', 'spglib',
             'ase_ext', 'flask', 'psycopg2', 'pyamg']
    return [format_dependency(name) for name in names]



# type: ignore
import os
import numpy as np
import numpy.testing
import unittest

import ase
import ase.build
import ase.io
from ase.io.vasp import write_vasp_xdatcar
from ase.calculators.calculator import compare_atoms


class TestXdatcarRoundtrip(unittest.TestCase):
    def setUp(self):
        self.outfile = 'NaCl.XDATCAR'

        self.NaCl = ase.build.bulk('NaCl', 'rocksalt', a=5.64)

    def tearDown(self):
        if os.path.isfile(self.outfile):
            os.remove(self.outfile)

    def assert_atoms_almost_equal(self, atoms, other, tol=1e-15):
        """Compare two Atoms objects, raising AssertionError if different"""
        system_changes = compare_atoms(atoms, other, tol=tol)

        if len(system_changes) > 0:
            raise AssertionError(
                "Atoms objects differ by {}".format(', '.join(system_changes)))

    def assert_trajectory_almost_equal(self, traj1, traj2):
        self.assertEqual(len(traj1), len(traj2))
        for image, other in zip(traj1, traj2):
            self.assert_atoms_almost_equal(image, other)

    def test_roundtrip(self):
        # Create a series of translated cells
        trajectory = [self.NaCl.copy() for i in range(5)]
        for i, atoms in enumerate(trajectory):
            atoms.set_scaled_positions(atoms.get_scaled_positions()
                                       + i * np.array([0.05, 0, 0.02]))
            atoms.wrap()

        ase.io.write(self.outfile, trajectory, format='vasp-xdatcar')
        roundtrip_trajectory = ase.io.read(self.outfile, index=':')
        self.assert_trajectory_almost_equal(trajectory, roundtrip_trajectory)

    def test_roundtrip_single_atoms(self):
        atoms = ase.build.bulk('Ge')
        ase.io.write(self.outfile, atoms, format='vasp-xdatcar')
        roundtrip_atoms = ase.io.read(self.outfile)
        self.assert_atoms_almost_equal(atoms, roundtrip_atoms)

    def test_typeerror(self):
        with self.assertRaises(TypeError):
            atoms = ase.build.bulk('Ge')
            write_vasp_xdatcar(self.outfile, atoms)
        with self.assertRaises(TypeError):
            not_atoms = 1
            ase.io.write(self.outfile, not_atoms, format='vasp-xdatcar')
        with self.assertRaises(TypeError):
            not_traj = [True, False, False]
            ase.io.write(self.outfile, not_traj, format='vasp-xdatcar')


def test_wrap():
    atoms = ase.build.bulk('Ge')
    # Shift atomic positions to get negative coordinates
    atoms.wrap(center=(-1, -1, -1))

    atoms.write('POSCAR', direct=True, wrap=False)
    new_atoms = ase.io.read('POSCAR')
    assert np.allclose(atoms.positions, new_atoms.positions)

    atoms.write('POSCAR', direct=True, wrap=True)
    new_atoms = ase.io.read('POSCAR')
    atoms.wrap()
    assert np.allclose(atoms.positions, new_atoms.positions)



import pytest
from ase.build import molecule
from ase.calculators.calculator import get_calculator_class
from ase.units import Ry
from ase.utils import workdir


# XXX To be replaced by stuff in ase.test.factories
class CalculatorInputs:
    def __init__(self, name, parameters=None):
        self.name = name
        if parameters is None:
            parameters = {}
        self.parameters = parameters

    def __repr__(self):
        cls = type(self)
        return '{}({}, {})'.format(cls.__name__,
                                   self.name, self.parameters)

    def calc(self):
        cls = get_calculator_class(self.name)
        return cls(**self.parameters)


def inputs(name, **parameters):
    return CalculatorInputs(name, parameters)


def _calculate(code, name):
    atoms = molecule(name)
    atoms.center(vacuum=3.5)
    with workdir('test-{}'.format(name), mkdir=True):
        atoms.calc = code.calc()
        return atoms.get_potential_energy()


@pytest.mark.parametrize(
    "spec",
    [
        inputs('gamess_us', label='ch4'),
        inputs('gaussian', xc='lda', basis='3-21G'),
    ],
    ids=lambda spec: spec.name)
def test_ch4(tmp_path, spec):
    # XXX Convert to string since pytest can sometimes gives us tmp_path
    # as a pathlib2 path.
    with workdir(str(tmp_path), mkdir=True):
        e_ch4 = _calculate(spec, 'CH4')
        e_c2h2 = _calculate(spec, 'C2H2')
        e_h2 = _calculate(spec, 'H2')
        energy = e_ch4 - 0.5 * e_c2h2 - 1.5 * e_h2
        print(energy)
        ref_energy = -2.8
        assert abs(energy - ref_energy) < 0.3


calc = pytest.mark.calculator
filterwarnings = pytest.mark.filterwarnings


@pytest.mark.calculator_lite
@calc('abinit', ecut=300, chksymbreak=0, toldfe=1e-4)
@calc('aims')
@calc('cp2k')
@calc('espresso', ecutwfc=300 / Ry)
@calc('gpaw', symmetry='off', mode='pw', txt='gpaw.txt', mixer={'beta': 0.6},
      marks=[filterwarnings('ignore:.*?ignore_bad_restart_file'),
             filterwarnings('ignore:convert_string_to_fd')])
@calc('nwchem')
@calc('octopus', Spacing='0.4 * angstrom')
@calc('openmx')
@calc('siesta', marks=pytest.mark.xfail)
def test_ch4_reaction(factory):
    e_ch4 = _calculate(factory, 'CH4')
    e_c2h2 = _calculate(factory, 'C2H2')
    e_h2 = _calculate(factory, 'H2')
    energy = e_ch4 - 0.5 * e_c2h2 - 1.5 * e_h2
    print(energy)
    ref_energy = -2.8
    assert abs(energy - ref_energy) < 0.3



def test_negativeindex():
    from ase.atoms import Atoms
    from ase.constraints import FixScaled

    a1 = Atoms(symbols='X2',
               positions=[[0., 0., 0.], [2., 0., 0.], ],
               cell=[[4., 0., 0.], [0., 4., 0.], [0., 0., 4.], ],
               )

    fs1 = FixScaled(a1.get_cell(), -1, mask=(True, False, False))
    fs2 = FixScaled(a1.get_cell(), 1, mask=(False, True, False))

    a1.set_constraint([fs1, fs2])

    # reassigning using atoms.__getitem__
    a2 = a1[0:2]

    assert len(a1._constraints) == len(a2._constraints)



def test_main():
    import numpy as np

    from ase.build import bulk
    from ase.constraints import FixScaled
    from ase.calculators.emt import EMT

    a = bulk("Ni", cubic=True)
    a.calc = EMT()

    pos = a.get_positions()

    a.set_constraint(
        FixScaled(
            a.cell,
            0,
        )
    )

    a.set_positions(
        pos * 1.01
    )

    assert np.sum(np.abs(a.get_forces()[0])) < 1e-12
    assert np.sum(np.abs(a.get_positions() - pos)[0]) < 1e-12
    assert np.sum(np.abs(a.get_positions() - pos*1.01)[1:].flatten()) < 1e-12



import ase.build


def test_siesta_read_eigenvalues(siesta_factory):
    # Test real calculation which produces a
    # gapped .EIG file
    atoms = ase.build.bulk('Si', cubic=True)
    calc = siesta_factory.calc(kpts=[2, 1, 1])
    atoms.calc = calc
    atoms.get_potential_energy()

    assert calc.results['eigenvalues'].shape[:2] == (1, 2)  # spins x bands
    assert calc.get_k_point_weights().shape == (2,)
    assert calc.get_ibz_k_points().shape == (2, 3)



import numpy as np
import pytest

from ase.units import GPa
from ase.build import bulk
from ase.calculators.test import gradient_test
from ase.constraints import UnitCellFilter, ExpCellFilter
from ase.optimize import LBFGS, MDMin
from ase.io import Trajectory


@pytest.fixture
def atoms(asap3):
    rng = np.random.RandomState(0)
    atoms = bulk('Cu', cubic=True)
    atoms.positions[:, 0] *= 0.995
    atoms.cell += rng.uniform(-1e-2, 1e-2, size=9).reshape((3, 3))
    atoms.calc = asap3.EMT()
    return atoms


@pytest.mark.parametrize('cellfilter', [UnitCellFilter, ExpCellFilter])
def test_pressure(atoms, cellfilter):
    xcellfilter = cellfilter(atoms, scalar_pressure=10.0 * GPa)

    # test all derivatives
    f, fn = gradient_test(xcellfilter)
    assert abs(f - fn).max() < 5e-6

    opt = LBFGS(xcellfilter)
    opt.run(1e-3)

    # check pressure is within 0.1 GPa of target
    sigma = atoms.get_stress() / GPa
    pressure = -(sigma[0] + sigma[1] + sigma[2]) / 3.0
    assert abs(pressure - 10.0) < 0.1


@pytest.mark.parametrize('cellfilter', [UnitCellFilter, ExpCellFilter])
def test_cellfilter(atoms, cellfilter):
    xcellfilter = cellfilter(atoms)
    f, fn = gradient_test(xcellfilter)
    assert abs(f - fn).max() < 3e-6


# XXX This test should have some assertions!  --askhl
def test_unitcellfilter(asap3, testdir):
    cu = bulk('Cu') * (6, 6, 6)
    cu.calc = asap3.EMT()
    f = UnitCellFilter(cu, [1, 1, 1, 0, 0, 0])
    opt = LBFGS(f)

    with Trajectory('Cu-fcc.traj', 'w', cu) as t:
        opt.attach(t)
        opt.run(5.0)
    # No assertions??


def test_unitcellfilter_hcp(asap3, testdir):
    cu = bulk('Cu', 'hcp', a=3.6 / 2.0**0.5)
    cu.cell[1, 0] -= 0.05
    cu *= (6, 6, 3)
    cu.calc = asap3.EMT()
    print(cu.get_forces())
    print(cu.get_stress())
    f = UnitCellFilter(cu)
    opt = MDMin(f, dt=0.01)
    with Trajectory('Cu-hcp.traj', 'w', cu) as t:
        opt.attach(t)
        opt.run(0.2)
    # No assertions??



# creates: ptable.png
import numpy as np
from ase import Atoms

# generates column, row positions for each element
zmax = 118
x, y = 1, 1  # column, row , initial coordinates for Hydrogen
positions = np.zeros((zmax + 1, 3))
for z in range(1, zmax + 1):
    if z == 2:
        x += 16
    if z == 5:
        x += 10
    if z == 13:
        x += 10
    if z == 57 or z == 89:
        y += 3
    if z == 72 or z == 104:
        y -= 3
        x -= 14
    positions[z] = (x, -y, 0)
    x += 1
    if x > 18:
        x = 1
        y += 1

spacing = 2.5
atoms = Atoms(np.arange(1, zmax + 1),
              positions[1:] * spacing)

atoms.write('ptable.png')

# from ase.visualize import view
# view(atoms)



import pytest
from ase.build import molecule
from numpy.testing import assert_allclose


@pytest.fixture()
def atoms():
    return molecule('H2O')


@pytest.mark.calculator('nwchem')
@pytest.mark.parametrize(
    'theory,eref,forces,pbc,kwargs',
    [
        ['dft', -2051.9802410863354, True, False, dict(basis='3-21G')],
        ['scf', -2056.7877421222634, True, False, dict(basis='3-21G')],
        ['mp2', -2060.1413846247333, True, False, dict(basis='3-21G')],
        ['ccsd', -2060.3418911515882, False, False, dict(basis='3-21G')],
        ['tce', -2060.319141863451, False, False, dict(
            basis='3-21G',
            tce={'ccd': None}
        )],
        ['tddft', -2044.3908422254976, True, False, dict(
            basis='3-21G',
            tddft=dict(
                nroots=2,
                algorithm=1,
                notriplet=None,
                target=1,
                civecs=None,
                grad={'root': 1},
            )
        )],
        ['pspw', -465.1290581383751, False, True, dict()],
        ['band', -465.1290611316276, False, True, dict()],
        ['paw', -2065.6600649367365, False, True, dict()]
    ]
)
def test_nwchem(factory, atoms, theory, eref, forces, pbc, kwargs):
    calc = factory.calc(label=theory, theory=theory, **kwargs)
    if pbc:
        atoms.center(vacuum=2)
        atoms.pbc = True
    atoms.calc = calc
    assert_allclose(atoms.get_potential_energy(), eref, atol=1e-4, rtol=1e-4)
    if forces:
        assert_allclose(atoms.get_forces(),
                        calc.calculate_numerical_forces(atoms),
                        atol=1e-4, rtol=1e-4)



"""WSGI Flask-app for browsing a database.

::

    +---------------------+
    | layout.html         |
    | +-----------------+ |    +--------------+
    | | search.html     | |    | layout.html  |
    | |     +           | |    | +---------+  |
    | | table.html ----------->| |row.html |  |
    | |                 | |    | +---------+  |
    | +-----------------+ |    +--------------+
    +---------------------+

You can launch Flask's local webserver like this::

    $ ase db abc.db -w

or this::

    $ python3 -m ase.db.app abc.db

"""

import io
import sys
from typing import Dict, Any, Set
from pathlib import Path

from flask import Flask, render_template, request

from ase.db import connect
from ase.db.core import Database
from ase.formula import Formula
from ase.db.web import create_key_descriptions, Session
from ase.db.row import row2dct, AtomsRow
from ase.db.table import all_columns


root = Path(__file__).parent.parent.parent
app = Flask(__name__, template_folder=str(root))

projects: Dict[str, Dict[str, Any]] = {}


@app.route('/', defaults={'project_name': 'default'})
@app.route('/<project_name>')
@app.route('/<project_name>/')
def search(project_name: str):
    """Search page.

    Contains input form for database query and a table result rows.
    """
    if project_name == 'favicon.ico':
        return '', 204, []  # 204: "No content"
    session = Session(project_name)
    project = projects[project_name]
    return render_template(project['search_template'],
                           q=request.args.get('query', ''),
                           p=project,
                           session_id=session.id)


@app.route('/update/<int:sid>/<what>/<x>/')
def update(sid: int, what: str, x: str):
    """Update table of rows inside search page.

    ``what`` must be one of:

    * query: execute query in request.args (x not used)
    * limit: set number of rows to show to x
    * toggle: toggle column x
    * sort: sort after column x
    * page: show page x
    """
    session = Session.get(sid)
    project = projects[session.project_name]
    session.update(what, x, request.args, project)
    table = session.create_table(project['database'],
                                 project['uid_key'],
                                 keys=list(project['key_descriptions']))
    return render_template(project['table_template'],
                           t=table,
                           p=project,
                           s=session)


@app.route('/<project_name>/row/<uid>')
def row(project_name: str, uid: str):
    """Show details for one database row."""
    project = projects[project_name]
    uid_key = project['uid_key']
    row = project['database'].get('{uid_key}={uid}'
                                  .format(uid_key=uid_key, uid=uid))
    dct = project['row_to_dict_function'](row, project)
    return render_template(project['row_template'],
                           d=dct, row=row, p=project, uid=uid)


@app.route('/atoms/<project_name>/<int:id>/<type>')
def atoms(project_name: str, id: int, type: str):
    """Return atomic structure as cif, xyz or json."""
    row = projects[project_name]['database'].get(id=id)
    a = row.toatoms()
    if type == 'cif':
        b = io.BytesIO()
        a.pbc = True
        a.write(b, 'cif', wrap=False)
        return b.getvalue(), 200, []

    fd = io.StringIO()
    if type == 'xyz':
        a.write(fd, 'xyz')
    elif type == 'json':
        con = connect(fd, type='json')
        con.write(row,
                  data=row.get('data', {}),
                  **row.get('key_value_pairs', {}))
    else:
        1 / 0

    headers = [('Content-Disposition',
                'attachment; filename="{project_name}-{id}.{type}"'
                .format(project_name=project_name, id=id, type=type))]
    txt = fd.getvalue()
    return txt, 200, headers


@app.route('/gui/<int:id>')
def gui(id: int):
    """Pop ud ase gui window."""
    from ase.visualize import view
    atoms = projects['default']['database'].get_atoms(id)
    view(atoms)
    return '', 204, []


@app.route('/test')
def test():
    from pyjokes import get_joke as j
    return j()


@app.route('/robots.txt')
def robots():
    return ('User-agent: *\n'
            'Disallow: /\n'
            '\n'
            'User-agent: Baiduspider\n'
            'Disallow: /\n'
            '\n'
            'User-agent: SiteCheck-sitecrawl by Siteimprove.com\n'
            'Disallow: /\n',
            200)


def handle_query(args) -> str:
    """Converts request args to ase.db query string."""
    return args['query']


def row_to_dict(row: AtomsRow, project: Dict[str, Any]) -> Dict[str, Any]:
    """Convert row to dict for use in html template."""
    dct = row2dct(row, project['key_descriptions'])
    dct['formula'] = Formula(Formula(row.formula).format('abc')).format('html')
    return dct


def add_project(db: Database) -> None:
    """Add database to projects with name 'default'."""
    all_keys: Set[str] = set()
    for row in db.select(columns=['key_value_pairs'], include_data=False):
        all_keys.update(row._keys)

    key_descriptions = {key: (key, '', '') for key in all_keys}

    meta: Dict[str, Any] = db.metadata

    if 'key_descriptions' in meta:
        key_descriptions.update(meta['key_descriptions'])

    default_columns = meta.get('default_columns')
    if default_columns is None:
        default_columns = all_columns[:]

    projects['default'] = {
        'name': 'default',
        'title': meta.get('title', ''),
        'uid_key': 'id',
        'key_descriptions': create_key_descriptions(key_descriptions),
        'database': db,
        'row_to_dict_function': row_to_dict,
        'handle_query_function': handle_query,
        'default_columns': default_columns,
        'search_template': 'ase/db/templates/search.html',
        'row_template': 'ase/db/templates/row.html',
        'table_template': 'ase/db/templates/table.html'}


if __name__ == '__main__':
    db = connect(sys.argv[1])
    add_project(db)
    app.run(host='0.0.0.0', debug=True)



# flake8: noqa
import numpy as np
from ase.io import read
from numpy.linalg import norm


def test_parse_socketio(testdir):
    write_output_socketio()
    traj = read("aims.out", ":", format="aims-output")

    a1, a2 = traj[0], traj[1]
    f1, f2 = a1.get_forces(), a2.get_forces()
    s1, s2 = a1.get_stress(voigt=False), a2.get_stress(voigt=False)

    assert np.allclose(a1.positions[1, 0], 2.11313574)
    assert np.allclose(a2.positions[1, 0], 2.11313574)

    assert np.allclose(f1[0, 0], -0.108555415821635e-07)
    assert np.allclose(f2[1, 1], 0.167235616064691e-03)

    assert np.allclose(s1[0, 0], 0.00006913)
    assert np.allclose(s2[0, 0], -0.00032660)


def test_run(testdir):
    write_output()
    atoms = read("aims.out", format="aims-output")

    # find total energy in aims.out
    key = "| Total energy corrected        :"
    with open("aims.out") as fd:
        line = next(l for l in fd if key in l)
        ref_energy = float(line.split()[5])

    assert norm(atoms.get_total_energy() - ref_energy) < 1e-12

    # find force in aims.out
    key = "Total atomic forces (unitary forces cleaned) [eV/Ang]:"
    with open("aims.out") as fd:
        next(l for l in fd if key in l)
        line = next(fd)
        ref_force = [float(l) for l in line.split()[2:5]]

    assert norm(atoms.get_forces()[0] - ref_force) < 1e-12

    # find stress in aims.out
    key = "Analytical stress tensor - Symmetrized"
    with open("aims.out") as fd:
        next(l for l in fd if key in l)
        # scroll to significant lines
        for _ in range(4):
            next(fd)
        line = next(fd)
        ref_stress = [float(l) for l in line.split()[2:5]]

    assert norm(atoms.get_stress(voigt=False)[0] - ref_stress) < 1e-12

    # find atomic stress in aims.out
    key = "Per atom stress (eV) used for heat flux calculation"
    with open("aims.out") as fd:
        next(l for l in fd if key in l)
        # scroll to boundary
        next(l for l in fd if "-------------" in l)

        line = next(fd)
        xx, yy, zz, xy, xz, yz = [float(l) for l in line.split()[2:8]]
        ref_stresses = [[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]]

    assert norm(atoms.get_stresses(voigt=False)[0] - ref_stresses) < 1e-12


def write_output():
    output = "  Basic array size parameters:\n  | Number of species                 :        1\n  | Number of atoms                   :        8\n  | Number of lattice vectors         :        3\n  | Max. basis fn. angular momentum   :        2\n  | Max. atomic/ionic basis occupied n:        3\n  | Max. number of basis fn. types    :        3\n  | Max. radial fns per species/type  :        5\n  | Max. logarithmic grid size        :     1346\n  | Max. radial integration grid size :       42\n  | Max. angular integration grid size:      302\n  | Max. angular grid division number :        8\n  | Radial grid for Hartree potential :     1346\n  | Number of spin channels           :        1\n\n\n  Input geometry:\n  | Unit cell:\n  |        5.42606753        0.00000000        0.00000000\n  |        0.00000000        5.42606753        0.00000000\n  |        0.00000000        0.00000000        5.42606753\n  | Atomic structure:\n  |       Atom                x [A]            y [A]            z [A]\n  |    1: Species Si            0.03431851       -0.09796859        0.09930953\n  |    2: Species Si            5.44231311        2.73920529        2.78205416\n  |    3: Species Si            2.75321969        0.10000784        2.72715717\n  |    4: Species Si            2.73199531        2.68826367       -0.08575931\n  |    5: Species Si            1.34757448        1.42946424        1.25761431\n  |    6: Species Si            1.35486030        4.13154987        4.06589071\n  |    7: Species Si            4.04177845        1.27675199        4.00805480\n  |    8: Species Si            3.99821025        4.01092826        1.42388121\n\n  +-------------------------------------------------------------------+\n  |              Analytical stress tensor - Symmetrized               |\n  |                  Cartesian components [eV/A**3]                   |\n  +-------------------------------------------------------------------+\n  |                x                y                z                |\n  |                                                                   |\n  |  x        -0.01478211      -0.01327277      -0.00355870           |\n  |  y        -0.01327277      -0.01512112      -0.01367280           |\n  |  z        -0.00355870      -0.01367280      -0.01534158           |\n  |                                                                   |\n  |  Pressure:       0.01508160   [eV/A**3]                           |\n  |                                                                   |\n  +-------------------------------------------------------------------+\n\n  ESTIMATED overall HOMO-LUMO gap:      0.21466369 eV between HOMO at k-point 1 and LUMO at k-point 1\n\n  Energy and forces in a compact form:\n  | Total energy uncorrected      :         -0.630943948216411E+05 eV\n  | Total energy corrected        :         -0.630943948568205E+05 eV  <-- do not rely on this value for anything but (periodic) metals\n  | Electronic free energy        :         -0.630943948919999E+05 eV\n  Total atomic forces (unitary forces cleaned) [eV/Ang]:\n  |   1         -0.104637839735875E+01          0.500412824184706E+00         -0.439789552504239E+00\n  |   2         -0.155820611394662E+00         -0.476557335046913E+00         -0.655396151432312E+00\n  |   3         -0.193381405004926E+01         -0.122454085397628E+01         -0.169259060410046E+01\n  |   4          0.404969041951871E-01          0.457139849737633E+00         -0.128445757910440E+00\n  |   5          0.109984435024380E-01         -0.165609149153507E+00          0.114351292468512E+01\n  |   6          0.663029766776301E+00         -0.814079627100908E-01          0.384378715376525E-04\n  |   7          0.213211510059627E+01          0.918575437083381E+00          0.189666102862743E+01\n  |   8          0.289372843732474E+00          0.719871898810707E-01         -0.123990325236629E+00\n\n\n    - Per atom stress (eV) used for heat flux calculation:\n        Atom   | Stress components (1,1), (2,2), (3,3), (1,2), (1,3), (2,3)\n      -------------------------------------------------------------------\n             1 |     0.9843662637E-01   -0.1027274769E+00    0.7237959330E-01   -0.3532042840E+00    0.2563317062E+00   -0.3642257991E+00\n             2 |     0.1244911861E+00   -0.4107147872E-01   -0.1084329966E+00    0.1201650287E+00   -0.1716383020E+00   -0.4669712541E-01\n             3 |    -0.1019986539E+01   -0.7054557814E+00   -0.8410240482E+00   -0.3714228752E+00   -0.4921256188E+00   -0.7970402772E+00\n             4 |    -0.5372048581E+00   -0.2498902919E+00   -0.2260340202E+00   -0.4368600591E+00    0.8622059429E-01    0.9182206824E-01\n             5 |    -0.3268304136E-01   -0.1853638313E+00    0.8046857169E-01   -0.3825550863E+00    0.3088175411E+00   -0.2399437437E+00\n             6 |    -0.2682129292E+00   -0.3832959470E+00   -0.5895171406E+00   -0.8151368635E-02    0.5046578049E-01   -0.6756388823E+00\n             7 |    -0.6970248515E+00   -0.6819450154E+00   -0.9123466446E+00   -0.5254451278E+00   -0.5070403877E+00   -0.6281674944E+00\n             8 |    -0.2933806554E-01   -0.6593089867E-01    0.7360641037E-01   -0.1629233327E+00   -0.9955320981E-01    0.4755870988E+00\n      -------------------------------------------------------------------\n\n\n          Have a nice day.\n------------------------------------------------------------\n"

    with open("aims.out", "w") as fd:
        fd.write(output)


def write_output_socketio():
    output = """
  MPI-parallelism will be employed.
------------------------------------------------------------
          Invoking FHI-aims ...

          When using FHI-aims, please cite the following reference:

          Volker Blum, Ralf Gehrke, Felix Hanke, Paula Havu,
          Ville Havu, Xinguo Ren, Karsten Reuter, and Matthias Scheffler,
          'Ab Initio Molecular Simulations with Numeric Atom-Centered Orbitals',
          Computer Physics Communications 180, 2175-2196 (2009)

          For any questions about FHI-aims, please visit the aimsclub website
          with its forums and wiki. Contributions to both the forums and the
          wiki are warmly encouraged - they are for you, and everyone is welcome there.

------------------------------------------------------------



  Date     :  20200311, Time     :  204314.253
  Time zero on CPU 1             :   0.194170000000000E-01  s.
  Internal wall clock time zero  :           353191394.253  s.

  FHI-aims created a unique identifier for this run for later identification
  aims_uuid : CF4170FC-8FB7-4BC2-A03C-9EC1F9140948

  Build configuration of the current instance of FHI-aims
  -------------------------------------------------------
  FHI-aims version      : 190813
  Commit number         : 6bedc122e
  CMake host system     : Linux-4.15.0-58-generic
  CMake version         : 3.5.1
  Fortran compiler      : /home/knoop/local/openmpi-3.0.0/build/bin/mpifort (Intel) version 17.0.1.20161005
  Fortran compiler flags: -O3 -ip -fp-model precise
  C compiler            : /usr/bin/gcc (GNU) version 8.3.0
  C compiler flags      :
  Architecture          :
  Using MPI
  Using C files
  Using LibXC
  Using SPGlib
  Using i-PI
  Linking against: /home/knoop/local/anaconda3/pkgs/mkl-2019.4-243/lib/libmkl_intel_lp64.so
                   /home/knoop/local/anaconda3/pkgs/mkl-2019.4-243/lib/libmkl_sequential.so
                   /home/knoop/local/anaconda3/pkgs/mkl-2019.4-243/lib/libmkl_core.so

  #===============================================================================
  # FHI-aims file: ./control.in
  # Created using the Atomic Simulation Environment (ASE)
  # Wed Mar 11 20:43:14 2020
  #===============================================================================
  xc                                 pbesol
  sc_accuracy_rho                    0.001
  compute_analytical_stress          .true.
  relativistic                       atomic_zora scalar
  output_level                       MD_light
  k_grid                             4 4 4
  use_pimd_wrapper                   localhost 12345
  #===============================================================================

  ################################################################################
  #
  #  FHI-aims code project
  #  VB, Fritz-Haber Institut, 2009
  #
  #  Suggested "light" defaults for Mg atom (to be pasted into control.in file)
  #  Be sure to double-check any results obtained with these settings for post-processing,
  #  e.g., with the "tight" defaults and larger basis sets.
  #
  ################################################################################
    species        Mg
  #     global species definitions
      nucleus             12
      mass                24.3050
  #
      l_hartree           4
  #
      cut_pot             4.0          1.5  1.0
      basis_dep_cutoff    1e-4
  #
      radial_base         40 5.5
      radial_multiplier   1
      angular_grids       specified
        division   0.7029   50
        division   0.9689  110
        division   1.1879  194
        division   1.3129  302
  #      division   1.4867  434
  #      division   1.6018  590
  #      division   1.8611  770
  #      division   1.9576  974
  #      division   2.2261 1202
  #      outer_grid   974
        outer_grid   302
  ################################################################################
  #
  #  Definition of "minimal" basis
  #
  ################################################################################
  #     valence basis states
      valence      3  s   2.
      valence      2  p   6.
  #     ion occupancy
      ion_occ      2  s   2.
      ion_occ      2  p   6.
  ################################################################################
  #
  #  Suggested additional basis functions. For production calculations,
  #  uncomment them one after another (the most important basis functions are
  #  listed first).
  #
  #  Constructed for dimers: 2.125 A, 2.375 A, 2.875 A, 3.375 A, 4.5 A
  #
  ################################################################################
  #  "First tier" - improvements: -230.76 meV to -21.94 meV
       hydro 2 p 1.5
       ionic 3 d auto
       hydro 3 s 2.4
  #  "Second tier" - improvements: -5.43 meV to -1.64 meV
  #     hydro 4 f 4.3
  #     hydro 2 p 3.4
  #     hydro 4 s 11.2
  #     hydro 3 d 6.2
  #  "Third tier" - improvements: -0.92 meV to -0.22 meV
  #     hydro 2 s 0.6
  #     hydro 3 p 4.8
  #     hydro 4 f 7.4
  #     hydro 5 g 6.6
  #     hydro 2 p 1.6
  #     hydro 3 d 1.8
  #  "Fourth tier" - improvements: -0.09 meV to -0.05 meV
  #     hydro 4 p 0.45
  #     hydro 5 g 10.4
  #     hydro 2 s 12.4
  #     hydro 4 d 1.7
  ################################################################################
  #
  #  FHI-aims code project
  #  VB, Fritz-Haber Institut, 2009
  #
  #  Suggested "light" defaults for O atom (to be pasted into control.in file)
  #  Be sure to double-check any results obtained with these settings for post-processing,
  #  e.g., with the "tight" defaults and larger basis sets.
  #
  ################################################################################
    species        O
  #     global species definitions
      nucleus             8
      mass                15.9994
  #
      l_hartree           4
  #
      cut_pot             3.5  1.5  1.0
      basis_dep_cutoff    1e-4
  #
      radial_base         36 5.0
      radial_multiplier   1
       angular_grids specified
        division   0.2659   50
        division   0.4451  110
        division   0.6052  194
        division   0.7543  302
  #      division   0.8014  434
  #      division   0.8507  590
  #      division   0.8762  770
  #      division   0.9023  974
  #      division   1.2339 1202
  #      outer_grid 974
        outer_grid 302
  ################################################################################
  #
  #  Definition of "minimal" basis
  #
  ################################################################################
  #     valence basis states
      valence      2  s   2.
      valence      2  p   4.
  #     ion occupancy
      ion_occ      2  s   1.
      ion_occ      2  p   3.
  ################################################################################
  #
  #  Suggested additional basis functions. For production calculations,
  #  uncomment them one after another (the most important basis functions are
  #  listed first).
  #
  #  Constructed for dimers: 1.0 A, 1.208 A, 1.5 A, 2.0 A, 3.0 A
  #
  ################################################################################
  #  "First tier" - improvements: -699.05 meV to -159.38 meV
       hydro 2 p 1.8
       hydro 3 d 7.6
       hydro 3 s 6.4
  #  "Second tier" - improvements: -49.91 meV to -5.39 meV
  #     hydro 4 f 11.6
  #     hydro 3 p 6.2
  #     hydro 3 d 5.6
  #     hydro 5 g 17.6
  #     hydro 1 s 0.75
  #  "Third tier" - improvements: -2.83 meV to -0.50 meV
  #     ionic 2 p auto
  #     hydro 4 f 10.8
  #     hydro 4 d 4.7
  #     hydro 2 s 6.8
  #  "Fourth tier" - improvements: -0.40 meV to -0.12 meV
  #     hydro 3 p 5
  #     hydro 3 s 3.3
  #     hydro 5 g 15.6
  #     hydro 4 f 17.6
  #     hydro 4 d 14
  # Further basis functions - -0.08 meV and below
  #     hydro 3 s 2.1
  #     hydro 4 d 11.6
  #     hydro 3 p 16
  #     hydro 2 s 17.2

  -----------------------------------------------------------------------
  Completed first pass over input file control.in .
  -----------------------------------------------------------------------


  -----------------------------------------------------------------------
  Parsing geometry.in (first pass over file, find array dimensions only).
  The contents of geometry.in will be repeated verbatim below
  unless switched off by setting 'verbatim_writeout .false.' .
  in the first line of geometry.in .
  -----------------------------------------------------------------------

  #=======================================================
  # FHI-aims file: ./geometry.in
  # Created using the Atomic Simulation Environment (ASE)
  # Wed Mar 11 20:43:14 2020
  #=======================================================
  lattice_vector 0.0000000000000000 2.1131357400000002 2.1131357400000002
  lattice_vector 2.1131357400000002 0.0000000000000000 2.1131357400000002
  lattice_vector 2.1131357400000002 2.1131357400000002 0.0000000000000000
  atom_frac 0.0000000000000000 0.0000000000000000 -0.0000000000000000 Mg
  atom_frac 0.5200000000000000 0.5000000000000000 0.5000000000000000 O

  -----------------------------------------------------------------------
  Completed first pass over input file geometry.in .
  -----------------------------------------------------------------------


  Basic array size parameters:
  | Number of species                 :        2
  | Number of atoms                   :        2
  | Number of lattice vectors         :        3
  | Max. basis fn. angular momentum   :        2
  | Max. atomic/ionic basis occupied n:        3
  | Max. number of basis fn. types    :        3
  | Max. radial fns per species/type  :        4
  | Max. logarithmic grid size        :     1334
  | Max. radial integration grid size :       40
  | Max. angular integration grid size:      302
  | Max. angular grid division number :        8
  | Radial grid for Hartree potential :     1334
  | Number of spin channels           :        1

------------------------------------------------------------
          Reading file control.in.
------------------------------------------------------------
  XC: Using PBEsol gradient-corrected functionals.
  Convergence accuracy of self-consistent charge density:  0.1000E-02
  Scalar relativistic treatment of kinetic energy: on-site free-atom approximation to ZORA.
  Requested output level: MD_light
  Found k-point grid:         4         4         4
  Using external wrapper (i-PI) for performing (path integral) molecular dynamics
  **Attention: initial geometry.in file will be ignored!

  Reading configuration options for species Mg                  .
  | Found nuclear charge :  12.0000
  | Found atomic mass :    24.3050000000000      amu
  | Found l_max for Hartree potential  :   4
  | Found cutoff potl. onset [A], width [A], scale factor :    4.00000    1.50000    1.00000
  | Threshold for basis-dependent cutoff potential is   0.100000E-03
  | Found data for basic radial integration grid :    40 points, outermost radius =    5.500 A
  | Found multiplier for basic radial grid :   1
  | Found angular grid specification: user-specified.
  | Specified grid contains     5 separate shells.
  | Check grid settings after all constraints further below.
  | Found free-atom valence shell :  3 s   2.000
  | Found free-atom valence shell :  2 p   6.000
  | Found free-ion valence shell :  2 s   2.000
  | Found free-ion valence shell :  2 p   6.000
  | Found hydrogenic basis function :  2 p   1.500
  | Found ionic basis function :  3 d , default cutoff radius.
  | Found hydrogenic basis function :  3 s   2.400
  Species Mg                  : Missing cutoff potential type.
  Defaulting to exp(1/x)/(1-x)^2 type cutoff potential.
  Species Mg: No 'logarithmic' tag. Using default grid for free atom:
  | Default logarithmic grid data [bohr] : 0.1000E-03 0.1000E+03 0.1012E+01
  | Will include ionic basis functions of  2.0-fold positive Mg                   ion.
  Species Mg: On-site basis accuracy parameter (for Gram-Schmidt orthonormalisation) not specified.
  Using default value basis_acc =  0.1000000E-03.
  Species Mg                  : Using default innermost maximum threshold i_radial=  2 for radial functions.
  Species Mg                  : Default cutoff onset for free atom density etc. : 0.40000000E+01 AA.
  Species Mg                  : Basic radial grid will be enhanced according to radial_multiplier =   1, to contain    40 grid points.

  Reading configuration options for species O                   .
  | Found nuclear charge :   8.0000
  | Found atomic mass :    15.9994000000000      amu
  | Found l_max for Hartree potential  :   4
  | Found cutoff potl. onset [A], width [A], scale factor :    3.50000    1.50000    1.00000
  | Threshold for basis-dependent cutoff potential is   0.100000E-03
  | Found data for basic radial integration grid :    36 points, outermost radius =    5.000 A
  | Found multiplier for basic radial grid :   1
  | Found angular grid specification: user-specified.
  | Specified grid contains     5 separate shells.
  | Check grid settings after all constraints further below.
  | Found free-atom valence shell :  2 s   2.000
  | Found free-atom valence shell :  2 p   4.000
  | Found free-ion valence shell :  2 s   1.000
  | Found free-ion valence shell :  2 p   3.000
  | Found hydrogenic basis function :  2 p   1.800
  | Found hydrogenic basis function :  3 d   7.600
  | Found hydrogenic basis function :  3 s   6.400
  Species O                   : Missing cutoff potential type.
  Defaulting to exp(1/x)/(1-x)^2 type cutoff potential.
  Species O : No 'logarithmic' tag. Using default grid for free atom:
  | Default logarithmic grid data [bohr] : 0.1000E-03 0.1000E+03 0.1012E+01
  Species O : On-site basis accuracy parameter (for Gram-Schmidt orthonormalisation) not specified.
  Using default value basis_acc =  0.1000000E-03.
  Species O                   : Using default innermost maximum threshold i_radial=  2 for radial functions.
  Species O                   : Default cutoff onset for free atom density etc. : 0.35000000E+01 AA.
  Species O                   : Basic radial grid will be enhanced according to radial_multiplier =   1, to contain    36 grid points.

  Finished reading input file 'control.in'. Consistency checks are next.

  MPI_IN_PLACE appears to work with this MPI implementation.
  | Keeping use_mpi_in_place .true. (see manual).
  Target number of points in a grid batch is not set. Defaulting to  100
  Method for grid partitioning is not set. Defaulting to parallel hash+maxmin partitioning.
  Batch size limit is not set. Defaulting to    200
  By default, will store active basis functions for each batch.
  If in need of memory, prune_basis_once .false. can be used to disable this option.
  communication_type for Hartree potential was not specified.
  Defaulting to calc_hartree .
  Defaulting to Pulay charge density mixer.
  Pulay mixer: Number of relevant iterations not set.
  Defaulting to    8 iterations.
  Pulay mixer: Number of initial linear mixing iterations not set.
  Defaulting to    0 iterations.
  Work space size for distributed Hartree potential not set.
  Defaulting to   0.200000E+03 MB.
  Mixing parameter for charge density mixing has not been set.
  Using default: charge_mix_param =     0.0500.
  The mixing parameter will be adjusted in iteration number     2 of the first full s.c.f. cycle only.
  Algorithm-dependent basis array size parameters:
  | n_max_pulay                         :        8
  Maximum number of self-consistency iterations not provided.
  Presetting  1000 iterations.
  Presetting      1001 iterations before the initial mixing cycle
  is restarted anyway using the sc_init_iter criterion / keyword.
  Presetting a factor      1.000 between actual scf density residual
  and density convergence criterion sc_accuracy_rho below which sc_init_iter
  takes no effect.
  Calculation of forces was not defined in control.in. No forces will be calculated.
  Geometry relaxation not requested: no relaxation will be performed.
  Analytical stress will be computed.
  Analytical stress calculation: Only the upper triangle is calculated.
                                 Final output is symmetrized.
  Analytical stress calculation: scf convergence accuracy of stress not set.
                                 Analytical stress self-consistency will not be checked explicitly.
                                 Be sure to set other criteria like sc_accuracy_rho tight enough.
  Force calculation: scf convergence accuracy of forces not set.
  Defaulting to 'sc_accuracy_forces not_checked'.
  Handling of forces: Unphysical translation and rotation will be removed from forces.
  No accuracy limit for integral partition fn. given. Defaulting to  0.1000E-14.
  No threshold value for u(r) in integrations given. Defaulting to  0.1000E-05.
  No occupation type (smearing scheme) given. Defaulting to Gaussian broadening, width =  0.1000E-01 eV.
  The width will be adjusted in iteration number     2 of the first full s.c.f. cycle only.
  S.C.F. convergence parameters will be adjusted in iteration number     2 of the first full s.c.f. cycle only.
  No accuracy for occupation numbers given. Defaulting to  0.1000E-12.
  No threshold value for occupation numbers given. Defaulting to  0.0000E+00.
  No accuracy for fermi level given. Defaulting to  0.1000E-19.
  Maximum # of iterations to find E_F not set. Defaulting to  200.
  Preferred method for the eigenvalue solver ('KS_method') not specified in 'control.in'.
  Defaulting to serial version, LAPACK (via ELSI), since more k-points than CPUs available.
  Will not use alltoall communication since running on < 1024 CPUs.
  Threshold for basis singularities not set.
  Default threshold for basis singularities:  0.1000E-04
  partition_type (choice of integration weights) for integrals was not specified.
  | Using a version of the partition function of Stratmann and coworkers ('stratmann_sparse').
  | At each grid point, the set of atoms used to build the partition table is smoothly restricted to
  | only those atoms whose free-atom density would be non-zero at that grid point.
  Partitioning for Hartree potential was not defined. Using partition_type for integrals.
  | Adjusted default value of keyword multip_moments_threshold to:       0.10000000E-11
  | This value may affect high angular momentum components of the Hartree potential in periodic systems.
  Spin handling was not defined in control.in. Defaulting to unpolarized case.
  Angular momentum expansion for Kerker preconditioner not set explicitly.
  | Using default value of   0
  No explicit requirement for turning off preconditioner.
  | By default, it will be turned off when the charge convergence reaches
  | sc_accuracy_rho  =   0.100000E-02
  No special mixing parameter while Kerker preconditioner is on.
  Using default: charge_mix_param =     0.0500.
  No q(lm)/r^(l+1) cutoff set for long-range Hartree potential.
  | Using default value of  0.100000E-09 .
  | Verify using the multipole_threshold keyword.
  Defaulting to new monopole extrapolation.
  Density update method: automatic selection selected.
  Using density matrix based charge density update.
  Using density matrix based charge density update.
  Using packed matrix style: index .
  Defaulting to use time-reversal symmetry for k-point grid.
------------------------------------------------------------


------------------------------------------------------------
          Reading geometry description geometry.in.
------------------------------------------------------------
  Input structure read successfully.
  The structure contains        2 atoms,  and a total of         20.000 electrons.

  Input geometry:
  | Unit cell:
  |        0.00000000        2.11313574        2.11313574
  |        2.11313574        0.00000000        2.11313574
  |        2.11313574        2.11313574        0.00000000
  | Atomic structure:
  |       Atom                x [A]            y [A]            z [A]
  |    1: Species Mg            0.00000000        0.00000000        0.00000000
  |    2: Species O             2.11313574        2.15539845        2.15539845

  Lattice parameters for 3D lattice (in Angstroms) :     2.988425    2.988425    2.988425
  Angle(s) between unit vectors (in degrees)       :    60.000000   60.000000   60.000000

  | The smallest distance between any two atoms is         2.07130423 AA.
  | The first atom of this pair is atom number                      2 .
  | The second atom of this pair is atom number                     1 .
  | Wigner-Seitz cell of the first atom image           0     1     0 .
  | (The Wigner-Seitz cell of the second atom is 0 0 0  by definition.)

  Symmetry information by spglib:
  | Precision set to  0.1E-04
  | Number of Operations  : 4
  | Space group           : 44
  | International         : Imm2
  | Schoenflies           : C2v^20

  Quantities derived from the lattice vectors:
  | Reciprocal lattice vector 1: -1.486697  1.486697  1.486697
  | Reciprocal lattice vector 2:  1.486697 -1.486697  1.486697
  | Reciprocal lattice vector 3:  1.486697  1.486697 -1.486697
  | Unit cell volume                               :   0.188718E+02  A^3

  Range separation radius for Ewald summation (hartree_convergence_parameter):      2.90574234 bohr.

  Number of empty states per atom not set in control.in - providing a guess from actual geometry.
  | Total number of empty states used during s.c.f. cycle:        6
  If you use a very high smearing, use empty_states (per atom!) in control.in to increase this value.

  Structure-dependent array size parameters:
  | Maximum number of distinct radial functions  :       13
  | Maximum number of basis functions            :       29
  | Number of Kohn-Sham states (occupied + empty):       16
------------------------------------------------------------

************************** W A R N I N G *******************************
* You are using the PIMD wrapper. Specifications and positions         *
* in geometry.in will be IGNORED - all is received from the wrapper.   *
* Please make sure species are declared in the same order in           *
* geometry.in and wrapper input.                                       *
************************************************************************


------------------------------------------------------------
          Preparing all fixed parts of the calculation.
------------------------------------------------------------
  Determining machine precision:
    2.225073858507201E-308
  Setting up grids for atomic and cluster calculations.

  Creating wave function, potential, and density for free atoms.

  Species: Mg

  List of occupied orbitals and eigenvalues:
    n    l              occ      energy [Ha]    energy [eV]
    1    0           2.0000       -46.359170     -1261.4972
    2    0           2.0000        -2.927498       -79.6613
    3    0           2.0000        -0.167982        -4.5710
    2    1           6.0000        -1.706030       -46.4234


  Species: O

  List of occupied orbitals and eigenvalues:
    n    l              occ      energy [Ha]    energy [eV]
    1    0           2.0000       -18.926989      -515.0296
    2    0           2.0000        -0.880247       -23.9527
    2    1           4.0000        -0.331514        -9.0210

  Creating fixed part of basis set: Ionic, confined, hydrogenic.

  Mg                   ion:

  List of free ionic orbitals and eigenvalues:
    n    l      energy [Ha]    energy [eV]
    1    0       -47.122987     -1282.2817
    2    0        -3.674099       -99.9773
    2    1        -2.451441       -66.7071


  List of ionic basis orbitals and eigenvalues:
    n    l      energy [Ha]    energy [eV]    outer radius [A]
    3    2        -0.253797        -6.9062       5.100062


  Mg                   hydrogenic:

  List of hydrogenic basis orbitals:
    n    l      effective z      eigenvalue [eV]  inner max. [A]     outer max. [A]     outer radius [A]
    2    1         1.500000        -7.5955           1.395712           1.395712           5.100062
    3    0         2.400000        -8.0111           0.162319           2.734035           5.100062


  O                    hydrogenic:

  List of hydrogenic basis orbitals:
    n    l      effective z      eigenvalue [eV]  inner max. [A]     outer max. [A]     outer radius [A]
    2    1         1.800000       -10.9749           1.164242           1.164242           4.578029
    3    2         7.600000       -87.3180           0.624125           0.624125           3.251020
    3    0         6.400000       -61.9207           0.061167           1.081902           4.001998


  Adding cutoff potential to free-atom effective potential.
  Creating atomic-like basis functions for current effective potential.

  Species Mg                  :

  List of atomic basis orbitals and eigenvalues:
    n    l      energy [Ha]    energy [eV]    outer radius [A]
    1    0       -46.359170     -1261.4972       0.921046
    2    0        -2.927498       -79.6613       3.621735
    3    0        -0.167982        -4.5710       5.100062
    2    1        -1.706030       -46.4234       4.404175


  Species O                   :

  List of atomic basis orbitals and eigenvalues:
    n    l      energy [Ha]    energy [eV]    outer radius [A]
    1    0       -18.926989      -515.0296       1.415765
    2    0        -0.880247       -23.9527       4.413171
    2    1        -0.331514        -9.0210       4.522403

  Assembling full basis from fixed parts.
  | Species Mg :   atomic orbital   1 s accepted.
  | Species Mg :   atomic orbital   2 s accepted.
  | Species Mg :   atomic orbital   3 s accepted.
  | Species Mg :    hydro orbital   3 s accepted.
  | Species Mg :   atomic orbital   2 p accepted.
  | Species Mg :    hydro orbital   2 p accepted.
  | Species Mg :    ionic orbital   3 d accepted.
  | Species O :   atomic orbital   1 s accepted.
  | Species O :    hydro orbital   3 s accepted.
  | Species O :   atomic orbital   2 s accepted.
  | Species O :   atomic orbital   2 p accepted.
  | Species O :    hydro orbital   2 p accepted.
  | Species O :    hydro orbital   3 d accepted.

  Basis size parameters after reduction:
  | Total number of radial functions:       13
  | Total number of basis functions :       29

  Per-task memory consumption for arrays in subroutine allocate_ext:
  |           4.441052MB.
  Testing on-site integration grid accuracy.
  |  Species  Function  <phi|h_atom|phi> (log., in eV)  <phi|h_atom|phi> (rad., in eV)
           1        1              -1261.4972086064              -1261.4966102663
           1        2                -79.6612686792                -79.6612661850
           1        3                 -4.5766449847                 -4.5764463354
           1        4                  4.2856803862                  4.2813226631
           1        5                -46.4234362291                -46.4234361907
           1        6                 -0.5282558506                 -0.5281575826
           1        7                  4.0144889854                  4.0133736295
           2        8               -515.0295626295               -515.0294562738
           2        9                 15.1698434419                 15.1699322204
           2       10                -21.6038822678                -21.6039118105
           2       11                 -9.0211123393                 -9.0212703513
           2       12                  8.3047696391                  8.2854840999
           2       13                 45.8428042125                 45.8427461222

  Preparing densities etc. for the partition functions (integrals / Hartree potential).

  Preparations completed.
  max(cpu_time)          :      0.155 s.
  Wall clock time (cpu1) :      0.155 s.
------------------------------------------------------------

************************** W A R N I N G *******************************
* Skipping the SCF initialization for now - done inside wrapper      *
************************************************************************

  @ DRIVER MODE: Connecting to host:port localhost       12345
  @ DRIVER MODE: Message from server: STATUS
  @ DRIVER MODE: Message from server: POSDATA
  @ DRIVER MODE: Received positions

------------------------------------------------------------
          Begin self-consistency loop: Initialization.

          Date     :  20200311, Time     :  204314.465
------------------------------------------------------------

  Initializing index lists of integration centers etc. from given atomic structure:
  Mapping all atomic coordinates to central unit cell.

  Initializing the k-points
          Using symmetry for reducing the k-points
  | k-points reduced from:       64 to       36
  | Number of k-points                             :        36
  The eigenvectors in the calculations are COMPLEX.
  | K-points in task   0:         9
  | K-points in task   1:         9
  | K-points in task   2:         9
  | K-points in task   3:         9
  | Number of basis functions in the Hamiltonian integrals :      1946
  | Number of basis functions in a single unit cell        :        29
  | Number of centers in hartree potential         :      1017
  | Number of centers in hartree multipole         :       776
  | Number of centers in electron density summation:       530
  | Number of centers in basis integrals           :       582
  | Number of centers in integrals                 :       189
  | Number of centers in hamiltonian               :       580
  | Consuming        186 KiB for k_phase.
  | Number of super-cells (origin) [n_cells]                     :        2197
  | Number of super-cells (after PM_index) [n_cells]             :         332
  | Number of super-cells in hamiltonian [n_cells_in_hamiltonian]:         332
  | Size of matrix packed + index [n_hamiltonian_matrix_size] :       70918
  | Estimated reciprocal-space cutoff momentum G_max:         3.25739508 bohr^-1 .
  | Reciprocal lattice points for long-range Hartree potential:      64
  Partitioning the integration grid into batches with parallel hashing+maxmin method.
  | Number of batches:      191
  | Maximal batch size:     112
  | Minimal batch size:      54
  | Average batch size:      62.492
  | Standard deviation of batch sizes:       5.898

  Integration load balanced across     4 MPI tasks.
  Work distribution over tasks is as follows:
  Task     0 has       3012 integration points.
  Task     1 has       2986 integration points.
  Task     2 has       2985 integration points.
  Task     3 has       2953 integration points.
  Initializing partition tables, free-atom densities, potentials, etc. across the integration grid (initialize_grid_storage).
  | initialize_grid_storage: Actual outermost partition radius vs. multipole_radius_free
  | (-- VB: in principle, multipole_radius_free should be larger, hence this output)
  | Species        1: Confinement radius =              5.500000000000000 AA, multipole_radius_free =              5.555717568450569 AA.
  | Species        1: outer_partition_radius set to              5.555717568450569 AA .
  | Species        2: Confinement radius =              4.999999999999999 AA, multipole_radius_free =              5.048384829883283 AA.
  | Species        2: outer_partition_radius set to              5.048384829883283 AA .
  | The sparse table of interatomic distances needs       1970.16 kbyte instead of      2709.79 kbyte of memory.
  | Net number of integration points:    11936
  | of which are non-zero points    :     9118
  | Numerical average free-atom electrostatic potential    :    -17.34765164 eV
  Renormalizing the initial density to the exact electron count on the 3D integration grid.
  | Initial density: Formal number of electrons (from input files) :      20.0000000000
  | Integrated number of electrons on 3D grid     :      20.0026739480
  | Charge integration error                      :       0.0026739480
  | Normalization factor for density and gradient :       0.9998663205
  Obtaining max. number of non-zero basis functions in each batch (get_n_compute_maxes).
  | Maximal number of non-zero basis functions:      807 in task     0
  | Maximal number of non-zero basis functions:      785 in task     1
  | Maximal number of non-zero basis functions:      825 in task     2
  | Maximal number of non-zero basis functions:      821 in task     3
  Allocating        0.064 MB for KS_eigenvector_complex
  Integrating Hamiltonian matrix: batch-based integration.
  Time summed over all CPUs for integration: real work        1.226 s, elapsed        1.244 s
  Integrating overlap matrix.
  Time summed over all CPUs for integration: real work        0.906 s, elapsed        0.920 s
  Decreasing sparse matrix size:
   Tolerance:  9.999999824516700E-014
   Hamiltonian matrix
  | Array has    57094 nonzero elements out of    70918 elements
  | Sparsity factor is 0.195
   Overlap matrix
  | Array has    53782 nonzero elements out of    70918 elements
  | Sparsity factor is 0.242
  New size of hamiltonian matrix:       57179

  Updating Kohn-Sham eigenvalues and eigenvectors using ELSI and the (modified) LAPACK eigensolver.
  Singularity check in k-point 4, task 0 (analysis for other k-points/tasks may follow below):

  Obtaining occupation numbers and chemical potential using ELSI.
  | Chemical potential (Fermi level):    -9.39880369 eV
  Writing Kohn-Sham eigenvalues.
  K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  State    Occupation    Eigenvalue [Ha]    Eigenvalue [eV]
      1       2.00000         -46.246119        -1258.42094
      2       2.00000         -18.949515         -515.64253
      3       2.00000          -2.939241          -79.98081
      4       2.00000          -1.724501          -46.92607
      5       2.00000          -1.724472          -46.92526
      6       2.00000          -1.724438          -46.92436
      7       2.00000          -1.074630          -29.24216
      8       2.00000          -0.457085          -12.43791
      9       2.00000          -0.456823          -12.43079
     10       2.00000          -0.456820          -12.43071
     11       0.00000          -0.221801           -6.03551
     12       0.00000           0.221833            6.03638
     13       0.00000           0.222549            6.05588
     14       0.00000           0.222984            6.06771
     15       0.00000           0.425167           11.56937
     16       0.00000           0.425678           11.58330

  What follows are estimated values for band gap, HOMO, LUMO, etc.
  | They are estimated on a discrete k-point grid and not necessarily exact.
  | For converged numbers, create a DOS and/or band structure plot on a denser k-grid.

  Highest occupied state (VBM) at    -12.43071357 eV (relative to internal zero)
  | Occupation number:      2.00000000
  | K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  Lowest unoccupied state (CBM) at    -6.03551047 eV (relative to internal zero)
  | Occupation number:      0.00000000
  | K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  ESTIMATED overall HOMO-LUMO gap:      6.39520310 eV between HOMO at k-point 1 and LUMO at k-point 1
  | This appears to be a direct band gap.
  The gap value is above 0.2 eV. Unless you are using a very sparse k-point grid,
  this system is most likely an insulator or a semiconductor.
  Calculating total energy contributions from superposition of free atom densities.

  Total energy components:
  | Sum of eigenvalues            :        -151.77161455 Ha       -4129.91575979 eV
  | XC energy correction          :         -24.34440425 Ha        -662.44494434 eV
  | XC potential correction       :          31.70940006 Ha         862.85667742 eV
  | Free-atom electrostatic energy:        -130.88597416 Ha       -3561.58856859 eV
  | Hartree energy correction     :           0.00000000 Ha           0.00000000 eV
  | Entropy correction            :           0.00000000 Ha           0.00000000 eV
  | ---------------------------
  | Total energy                  :        -275.29259290 Ha       -7491.09259530 eV
  | Total energy, T -> 0          :        -275.29259290 Ha       -7491.09259530 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy        :        -275.29259290 Ha       -7491.09259530 eV

  Derived energy quantities:
  | Kinetic energy                :         276.03701172 Ha        7511.34926225 eV
  | Electrostatic energy          :        -526.98520038 Ha      -14339.99691322 eV
  | Energy correction for multipole
  | error in Hartree potential    :           0.00000000 Ha           0.00000000 eV
  | Sum of eigenvalues per atom                           :       -2064.95787990 eV
  | Total energy (T->0) per atom                          :       -3745.54629765 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy per atom                       :       -3745.54629765 eV
  Initialize hartree_potential_storage
  Max. number of atoms included in rho_multipole:            2

  End scf initialization - timings             :  max(cpu_time)    wall_clock(cpu1)
  | Time for scf. initialization               :        0.904 s           0.904 s
  | Boundary condition initialization          :        0.062 s           0.061 s
  | Integration                                :        0.544 s           0.544 s
  | Solution of K.-S. eqns.                    :        0.007 s           0.007 s
  | Grid partitioning                          :        0.016 s           0.016 s
  | Preloading free-atom quantities on grid    :        0.002 s           0.002 s
  | Free-atom superposition energy             :        0.023 s           0.023 s
  | Total energy evaluation                    :        0.000 s           0.000 s

  Partial memory accounting:
  | Current value for overall tracked memory usage:
  |   Minimum:        1.155 MB (on task 0)
  |   Maximum:        1.155 MB (on task 0)
  |   Average:        1.155 MB
  | Peak value for overall tracked memory usage:
  |   Minimum:        8.148 MB (on task 1 after allocating wave)
  |   Maximum:        8.743 MB (on task 2 after allocating wave)
  |   Average:        8.512 MB
  | Largest tracked array allocation so far:
  |   Minimum:        4.701 MB (hamiltonian_shell on task 1)
  |   Maximum:        5.193 MB (hamiltonian_shell on task 2)
  |   Average:        5.001 MB
  Note:  These values currently only include a subset of arrays which are explicitly tracked.
  The "true" memory usage will be greater.
------------------------------------------------------------
Convergence:    q app. |  density  | eigen (eV) | Etot (eV) | forces (eV/A) |       CPU time |     Clock time
  SCF    1 :  0.28E-02 |  0.27E+00 |   0.14E+01 |  0.25E+00 |             . |        1.078 s |        1.078 s
  SCF    2 :  0.42E-02 |  0.24E+00 |   0.14E+02 |  0.11E+01 |             . |        1.076 s |        1.076 s
  SCF    3 :  0.42E-02 |  0.81E-01 |  -0.54E+01 | -0.20E-01 |             . |        1.068 s |        1.068 s
  SCF    4 :  0.45E-02 |  0.66E-01 |  -0.33E+01 |  0.72E-01 |             . |        1.062 s |        1.063 s
  SCF    5 :  0.35E-02 |  0.32E-01 |   0.37E+01 |  0.15E-01 |             . |        1.069 s |        1.069 s
  SCF    6 :  0.34E-02 |  0.56E-02 |   0.32E-01 |  0.38E-03 |             . |        1.072 s |        1.072 s
  SCF    7 :  0.28E-02 |  0.26E-02 |   0.86E-01 | -0.15E-05 |             . |        1.077 s |        1.077 s
  SCF    8 :  0.27E-02 |  0.49E-03 |  -0.15E-01 |  0.48E-05 |             . |        1.133 s |        1.134 s
  SCF    9 :  0.27E-02 |  0.63E-03 |   0.33E-01 | -0.18E-04 |      0.11E+01 |        1.454 s |        1.453 s
  SCF   10 :  0.27E-02
  Analytical stress tensor components [eV]         xx                  yy                  zz                  xy                  xz                  yz
  -----------------------------------------------------------------------------------------------------------------------------------------------------------
    Nuclear Hellmann-Feynman      :    -0.3870932949E+02   -0.3875207926E+02   -0.3875207926E+02    0.4430429868E-10    0.4427216952E-10    0.2146800652E-01
    Multipole Hellmann-Feynman    :    -0.5289450565E+02   -0.5304899474E+02   -0.5304899474E+02    0.2987199479E-10    0.3011368558E-10    0.8812370048E-01
    On-site Multipole corrections :    -0.1380739258E+00   -0.1382119934E+00   -0.1382119933E+00   -0.1672464662E-10   -0.4833712897E-12    0.2520670526E-03
    Strain deriv. of the orbitals :     0.9174321372E+02    0.9186596802E+02    0.9186596802E+02    0.2265077863E-09    0.5002460457E-09   -0.9978486771E-01
  -----------------------------------------------------------------------------------------------------------------------------------------------------------
  Sum of all contributions        :     0.1304653355E-02   -0.7331797879E-01   -0.7331797897E-01    0.2839594332E-09    0.5741485295E-09    0.1005890634E-01

  +-------------------------------------------------------------------+
  |              Analytical stress tensor - Symmetrized               |
  |                  Cartesian components [eV/A**3]                   |
  +-------------------------------------------------------------------+
  |                x                y                z                |
  |                                                                   |
  |  x         0.00006913       0.00000000       0.00000000           |
  |  y         0.00000000      -0.00388507       0.00053301           |
  |  z         0.00000000       0.00053301      -0.00388507           |
  |                                                                   |
  |  Pressure:       0.00256700   [eV/A**3]                           |
  |                                                                   |
  +-------------------------------------------------------------------+

 * Warning: Stress tensor is anisotropic. Be aware that pressure is an isotropic quantity.

 |  0.12E-03 |   0.32E-01 | -0.12E-04 |      0.12E-02 |       14.861 s |       14.863 s

  Total energy components:
  | Sum of eigenvalues            :        -151.37923990 Ha       -4119.23870235 eV
  | XC energy correction          :         -24.44685738 Ha        -665.23283605 eV
  | XC potential correction       :          31.84308529 Ha         866.49443746 eV
  | Free-atom electrostatic energy:        -130.88597416 Ha       -3561.58856859 eV
  | Hartree energy correction     :          -0.37115708 Ha         -10.09969792 eV
  | Entropy correction            :           0.00000000 Ha           0.00000000 eV
  | ---------------------------
  | Total energy                  :        -275.24014324 Ha       -7489.66536745 eV
  | Total energy, T -> 0          :        -275.24014324 Ha       -7489.66536745 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy        :        -275.24014324 Ha       -7489.66536745 eV

  Derived energy quantities:
  | Kinetic energy                :         275.72011310 Ha        7502.72601198 eV
  | Electrostatic energy          :        -526.51339896 Ha      -14327.15854338 eV
  | Energy correction for multipole
  | error in Hartree potential    :           0.00356157 Ha           0.09691537 eV
  | Sum of eigenvalues per atom                           :       -2059.61935117 eV
  | Total energy (T->0) per atom                          :       -3744.83268372 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy per atom                       :       -3744.83268372 eV
  What follows are estimated values for band gap, HOMO, LUMO, etc.
  | They are estimated on a discrete k-point grid and not necessarily exact.
  | For converged numbers, create a DOS and/or band structure plot on a denser k-grid.

  Highest occupied state (VBM) at     -9.64325638 eV (relative to internal zero)
  | Occupation number:      2.00000000
  | K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  Lowest unoccupied state (CBM) at    -5.16542680 eV (relative to internal zero)
  | Occupation number:      0.00000000
  | K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  ESTIMATED overall HOMO-LUMO gap:      4.47782958 eV between HOMO at k-point 1 and LUMO at k-point 1
  | This appears to be a direct band gap.
  The gap value is above 0.2 eV. Unless you are using a very sparse k-point grid,
  this system is most likely an insulator or a semiconductor.

  Self-consistency cycle converged.

  Removing unitary transformations (pure translations, rotations) from forces on atoms.
  Atomic forces before filtering:
  | Net force on center of mass :  -0.379388E-08 -0.426476E-02 -0.426475E-02 eV/A
  Atomic forces after filtering:
  | Net force on center of mass :  -0.265846E-23 -0.446015E-16  0.000000E+00 eV/A

  Energy and forces in a compact form:
  | Total energy uncorrected      :         -0.748966536744980E+04 eV
  | Total energy corrected        :         -0.748966536744980E+04 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy        :         -0.748966536744980E+04 eV
  Total atomic forces (unitary forces cleaned) [eV/Ang]:
  |    1         -0.108555415821635E-07          0.232312241621526E+00          0.232312248124934E+00
  |    2          0.108555415821635E-07         -0.232312241621526E+00         -0.232312248124934E+00

  Removing unitary transformations (pure translations, rotations) from forces on atoms.
  Atomic forces before filtering:
  | Net force on center of mass :  -0.265846E-23 -0.446015E-16  0.000000E+00 eV/A
  Atomic forces after filtering:
  | Net force on center of mass :   0.000000E+00  0.446015E-16  0.000000E+00 eV/A
 ------------------------------------------------------------------------
 Atomic structure that was used in the preceding time step of the wrapper
                         x [A]             y [A]             z [A]
  lattice_vector         0.00000000        2.11313574        2.11313574
  lattice_vector         2.11313574        0.00000000        2.11313574
  lattice_vector         2.11313574        2.11313574        0.00000000

            atom         0.00000000        0.00000000        0.00000000  Mg
            atom         2.11313574        0.04226271        0.04226271  O
 ------------------------------------------------------------------------
  @ DRIVER MODE: Message from server: STATUS
  @ DRIVER MODE: Message from server: GETFORCE
  @ DRIVER MODE: Returning v,forces,stress
  @ DRIVER MODE: Message from server: STATUS
  @ DRIVER MODE: Message from server: INIT
  @ DRIVER MODE: Receiving replica           0
  @ DRIVER MODE: Message from server: STATUS
  @ DRIVER MODE: Message from server: POSDATA
  @ DRIVER MODE: Received positions

------------------------------------------------------------
          Begin self-consistency loop: Re-initialization.

  Date     :  20200311, Time     :  204340.453
------------------------------------------------------------

  Initializing index lists of integration centers etc. from given atomic structure:
  Mapping all atomic coordinates to central unit cell.

  Initializing the k-points
          Using symmetry for reducing the k-points
  | k-points reduced from:       64 to       36
  | Number of k-points                             :        36
  The eigenvectors in the calculations are COMPLEX.
  | K-points in task   0:         9
  | K-points in task   1:         9
  | K-points in task   2:         9
  | K-points in task   3:         9
  | Number of basis functions in the Hamiltonian integrals :      1949
  | Number of basis functions in a single unit cell        :        29
  | Number of centers in hartree potential         :      1097
  | Number of centers in hartree multipole         :       842
  | Number of centers in electron density summation:       585
  | Number of centers in basis integrals           :       633
  | Number of centers in integrals                 :       203
  | Number of centers in hamiltonian               :       654
  | Consuming        228 KiB for k_phase.
  | Number of super-cells (origin) [n_cells]                     :        2197
  | Number of super-cells (after PM_index) [n_cells]             :         406
  | Number of super-cells in hamiltonian [n_cells_in_hamiltonian]:         406
  | Size of matrix packed + index [n_hamiltonian_matrix_size] :       70804
  Partitioning the integration grid into batches with parallel hashing+maxmin method.
  Initializing partition tables, free-atom densities, potentials, etc. across the integration grid (initialize_grid_storage).
  | Species        1: outer_partition_radius set to              5.555717568450569 AA .
  | Species        2: outer_partition_radius set to              5.048384829883283 AA .
  | The sparse table of interatomic distances needs       2198.75 kbyte instead of      3205.51 kbyte of memory.
  | Net number of integration points:    11936
  | of which are non-zero points    :     9174
  Renormalizing the initial density to the exact electron count on the 3D integration grid.
  | Initial density: Formal number of electrons (from input files) :      20.0000000000
  | Integrated number of electrons on 3D grid     :      20.0033111186
  | Charge integration error                      :       0.0033111186
  | Normalization factor for density and gradient :       0.9998344715
  Calculating total energy contributions from superposition of free atom densities.
  Initialize hartree_potential_storage
  Max. number of atoms included in rho_multipole:            2
  Integrating overlap matrix.
  Time summed over all CPUs for integration: real work        1.006 s, elapsed        1.029 s
  Orthonormalizing eigenvectors

  End scf reinitialization - timings           :  max(cpu_time)    wall_clock(cpu1)
  | Time for scf. reinitialization             :        0.696 s           0.697 s
  | Boundary condition initialization          :        0.095 s           0.095 s
  | Integration                                :        0.258 s           0.258 s
  | Grid partitioning                          :        0.019 s           0.019 s
  | Preloading free-atom quantities on grid    :        0.296 s           0.295 s
  | Free-atom superposition energy             :        0.026 s           0.027 s
  | K.-S. eigenvector reorthonormalization     :        0.002 s           0.001 s
------------------------------------------------------------
Convergence:    q app. |  density  | eigen (eV) | Etot (eV) | forces (eV/A) |       CPU time |     Clock time
  SCF    1 :  0.33E-02 |  0.39E+00 |  -0.41E+04 | -0.75E+04 |             . |        1.072 s |        1.072 s
  SCF    2 :  0.32E-02 |  0.91E+00 |  -0.22E+02 |  0.79E+01 |             . |        1.400 s |        1.400 s
  SCF    3 :  0.31E-02 |  0.56E+00 |  -0.39E+02 |  0.45E+01 |             . |        1.449 s |        1.450 s
  SCF    4 :  0.31E-02 |  0.59E-01 |   0.28E+01 |  0.14E-01 |             . |        1.252 s |        1.253 s
  SCF    5 :  0.32E-02 |  0.25E-01 |  -0.32E+00 | -0.21E-02 |             . |        1.350 s |        1.349 s
  SCF    6 :  0.33E-02 |  0.32E-02 |  -0.25E+00 |  0.29E-03 |             . |        1.176 s |        1.175 s
  SCF    7 :  0.35E-02 |  0.49E-02 |   0.60E+00 |  0.44E-04 |             . |        1.185 s |        1.186 s
  SCF    8 :  0.34E-02 |  0.11E-02 |   0.10E+00 | -0.69E-04 |             . |        1.165 s |        1.165 s
  SCF    9 :  0.34E-02 |  0.82E-04 |   0.41E-02 | -0.33E-05 |             . |        1.168 s |        1.168 s
  SCF   10 :  0.33E-02 |  0.72E-04 |  -0.19E-02 | -0.80E-05 |      0.33E-02 |        1.642 s |        1.642 s
  SCF   11 :  0.33E-02
  Analytical stress tensor components [eV]         xx                  yy                  zz                  xy                  xz                  yz
  -----------------------------------------------------------------------------------------------------------------------------------------------------------
    Nuclear Hellmann-Feynman      :    -0.3864674379E+02   -0.3864415535E+02   -0.3864415535E+02   -0.2513001734E-11   -0.2404967012E-11    0.2659262711E-04
    Multipole Hellmann-Feynman    :    -0.5297314950E+02   -0.5297337267E+02   -0.5297337267E+02   -0.2417076525E-11   -0.1885368919E-11    0.4490875098E-05
    On-site Multipole corrections :    -0.1375921434E+00   -0.1375921435E+00   -0.1375921434E+00    0.1740136643E-11    0.5800455476E-11    0.4302004478E-11
    Strain deriv. of the orbitals :     0.9175132185E+02    0.9174867058E+02    0.9174867058E+02   -0.4651096734E-09   -0.3547552527E-09   -0.1085119856E-04
  -----------------------------------------------------------------------------------------------------------------------------------------------------------
  Sum of all contributions        :    -0.6163591118E-02   -0.6449573826E-02   -0.6449574530E-02   -0.4682996150E-09   -0.3532451331E-09    0.2023230795E-04

  +-------------------------------------------------------------------+
  |              Analytical stress tensor - Symmetrized               |
  |                  Cartesian components [eV/A**3]                   |
  +-------------------------------------------------------------------+
  |                x                y                z                |
  |                                                                   |
  |  x        -0.00032660      -0.00000000      -0.00000000           |
  |  y        -0.00000000      -0.00034176       0.00000107           |
  |  z        -0.00000000       0.00000107      -0.00034176           |
  |                                                                   |
  |  Pressure:       0.00033671   [eV/A**3]                           |
  |                                                                   |
  +-------------------------------------------------------------------+

 * Warning: Stress tensor is anisotropic. Be aware that pressure is an isotropic quantity.

 |  0.17E-04 |   0.99E-03 | -0.18E-05 |      0.29E-02 |       16.571 s |       16.572 s

  Total energy components:
  | Sum of eigenvalues            :        -151.38002476 Ha       -4119.26005928 eV
  | XC energy correction          :         -24.44607870 Ha        -665.21164709 eV
  | XC potential correction       :          31.84204127 Ha         866.46602823 eV
  | Free-atom electrostatic energy:        -130.88928774 Ha       -3561.67873571 eV
  | Hartree energy correction     :          -0.36705358 Ha          -9.98803622 eV
  | Entropy correction            :           0.00000000 Ha           0.00000000 eV
  | ---------------------------
  | Total energy                  :        -275.24040352 Ha       -7489.67245007 eV
  | Total energy, T -> 0          :        -275.24040352 Ha       -7489.67245007 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy        :        -275.24040352 Ha       -7489.67245007 eV

  Derived energy quantities:
  | Kinetic energy                :         275.71490212 Ha        7502.58421404 eV
  | Electrostatic energy          :        -526.50922694 Ha      -14327.04501702 eV
  | Energy correction for multipole
  | error in Hartree potential    :           0.00349067 Ha           0.09498603 eV
  | Sum of eigenvalues per atom                           :       -2059.63002964 eV
  | Total energy (T->0) per atom                          :       -3744.83622503 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy per atom                       :       -3744.83622503 eV
  What follows are estimated values for band gap, HOMO, LUMO, etc.
  | They are estimated on a discrete k-point grid and not necessarily exact.
  | For converged numbers, create a DOS and/or band structure plot on a denser k-grid.

  Highest occupied state (VBM) at     -9.63993423 eV (relative to internal zero)
  | Occupation number:      2.00000000
  | K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  Lowest unoccupied state (CBM) at    -5.16099556 eV (relative to internal zero)
  | Occupation number:      0.00000000
  | K-point:       1 at    0.000000    0.000000    0.000000 (in units of recip. lattice)

  ESTIMATED overall HOMO-LUMO gap:      4.47893868 eV between HOMO at k-point 1 and LUMO at k-point 1
  | This appears to be a direct band gap.
  The gap value is above 0.2 eV. Unless you are using a very sparse k-point grid,
  this system is most likely an insulator or a semiconductor.

  Self-consistency cycle converged.

  Removing unitary transformations (pure translations, rotations) from forces on atoms.
  Atomic forces before filtering:
  | Net force on center of mass :  -0.173882E-08  0.213870E-03  0.213872E-03 eV/A
  Atomic forces after filtering:
  | Net force on center of mass :  -0.132923E-23  0.000000E+00  0.653343E-19 eV/A

  Energy and forces in a compact form:
  | Total energy uncorrected      :         -0.748967245006648E+04 eV
  | Total energy corrected        :         -0.748967245006648E+04 eV  <-- do not rely on this value for anything but (periodic) metals
  | Electronic free energy        :         -0.748967245006648E+04 eV
  Total atomic forces (unitary forces cleaned) [eV/Ang]:
  |    1          0.773348206613320E-08         -0.167235616064691E-03         -0.167231728110871E-03
  |    2         -0.773348206613320E-08          0.167235616064691E-03          0.167231728110871E-03

  Removing unitary transformations (pure translations, rotations) from forces on atoms.
  Atomic forces before filtering:
  | Net force on center of mass :  -0.132923E-23  0.000000E+00  0.653343E-19 eV/A
  Atomic forces after filtering:
  | Net force on center of mass :  -0.132923E-23  0.000000E+00  0.217781E-19 eV/A
 ------------------------------------------------------------------------
 Atomic structure that was used in the preceding time step of the wrapper
                         x [A]             y [A]             z [A]
  lattice_vector         0.00000000        2.11313574        2.11313574
  lattice_vector         2.11313574        0.00000000        2.11313574
  lattice_vector         2.11313574        2.11313574        0.00000000

            atom         0.00000000        0.00000000        0.00000000  Mg
            atom         2.11313574        2.11313574        2.11313574  O
 ------------------------------------------------------------------------
  @ DRIVER MODE: Message from server: STATUS
  @ DRIVER MODE: Message from server: GETFORCE
  @ DRIVER MODE: Returning v,forces,stress

------------------------------------------------------------------------------
  Final output of selected total energy values:

  The following output summarizes some interesting total energy values
  at the end of a run (AFTER all relaxation, molecular dynamics, etc.).

  | Total energy of the DFT / Hartree-Fock s.c.f. calculation      :          -7489.672450066 eV
  | Final zero-broadening corrected energy (caution - metals only) :          -7489.672450066 eV
  | For reference only, the value of 1 Hartree used in FHI-aims is :             27.211384500 eV
  | For reference only, the overall average (free atom contribution
  | + realspace contribution) of the electrostatic potential after
  | s.c.f. convergence is                                          :            -16.786786172 eV

  Before relying on these values, please be sure to understand exactly which
  total energy value is referred to by a given number. Different objects may
  all carry the same name 'total energy'. Definitions:

  Total energy of the DFT / Hartree-Fock s.c.f. calculation:
  | Note that this energy does not include ANY quantities calculated after the
  | s.c.f. cycle, in particular not ANY RPA, MP2, etc. many-body perturbation terms.

  Final zero-broadening corrected energy:
  | For metallic systems only, a broadening of the occupation numbers at the Fermi
  | level can be extrapolated back to zero broadening by an electron-gas inspired
  | formula. For all systems that are not real metals, this value can be
  | meaningless and should be avoided.

------------------------------------------------------------------------------
  Methods described in the following list of references were used in this FHI-aims run.
  If you publish the results, please make sure to cite these reference if they apply.
  FHI-aims is an academic code, and for our developers (often, Ph.D. students
  and postdocs), scientific credit in the community is essential.
  Thank you for helping us!

  For any use of FHI-aims, please cite:

    Volker Blum, Ralf Gehrke, Felix Hanke, Paula Havu, Ville Havu,
    Xinguo Ren, Karsten Reuter, and Matthias Scheffler
    'Ab initio molecular simulations with numeric atom-centered orbitals'
    Computer Physics Communications 180, 2175-2196 (2009)
    http://dx.doi.org/10.1016/j.cpc.2009.06.022


  For the analytical stress tensor used in your run, please cite:

    Franz Knuth, Christian Carbogno, Viktor Atalla, Volker Blum, Matthias Scheffler
    'All-electron formalism for total energy strain derivatives and
    stress tensor components for numeric atom-centered orbitals'
    Computer Physics Communications 190, 33-50 (2015).
    http://dx.doi.org/10.1016/j.cpc.2015.01.003


  The provided symmetry information was generated with SPGlib:

    Atsushi Togo, Yusuke Seto, Dimitar Pashov
    SPGlib 1.7.3 obtained from http://spglib.sourceforge.net
    Copyright (C) 2008 Atsushi Togo


  The ELSI infrastructure was used in your run to solve the Kohn-Sham electronic structure.
  Please check out http://elsi-interchange.org to learn more.
  If scalability is important for your project, please acknowledge ELSI by citing:

    V. W-z. Yu, F. Corsetti, A. Garcia, W. P. Huhn, M. Jacquelin, W. Jia,
    B. Lange, L. Lin, J. Lu, W. Mi, A. Seifitokaldani, A. Vazquez-Mayagoitia,
    C. Yang, H. Yang, and V. Blum
    'ELSI: A unified software interface for Kohn-Sham electronic structure solvers'
    Computer Physics Communications 222, 267-285 (2018).
    http://dx.doi.org/10.1016/j.cpc.2017.09.007


  For the real-space grid partitioning and parallelization used in this calculation, please cite:

    Ville Havu, Volker Blum, Paula Havu, and Matthias Scheffler,
    'Efficient O(N) integration for all-electron electronic structure calculation'
    'using numerically tabulated basis functions'
    Journal of Computational Physics 228, 8367-8379 (2009).
    http://dx.doi.org/10.1016/j.jcp.2009.08.008

  Of course, there are many other important community references, e.g., those cited in the
  above references. Our list is limited to references that describe implementations in the
  FHI-aims code. The reason is purely practical (length of this list) - please credit others as well.

------------------------------------------------------------
          Leaving FHI-aims.
          Date     :  20200311, Time     :  204410.632

          Computational steps:
          | Number of self-consistency cycles          :           21
          | Number of SCF (re)initializations          :            2

          Detailed time accounting                     :  max(cpu_time)    wall_clock(cpu1)
          | Total time                                 :       56.355 s          56.379 s
          | Preparation time                           :        0.155 s           0.155 s
          | Boundary condition initalization           :        0.157 s           0.156 s
          | Grid partitioning                          :        0.035 s           0.035 s
          | Preloading free-atom quantities on grid    :        0.298 s           0.297 s
          | Free-atom superposition energy             :        0.049 s           0.050 s
          | Total time for integrations                :        8.101 s           8.102 s
          | Total time for solution of K.-S. equations :        0.136 s           0.136 s
          | Total time for density & force components  :       40.500 s          40.501 s
          | Total time for mixing & preconditioning    :        1.391 s           1.394 s
          | Total time for Hartree multipole update    :        0.013 s           0.012 s
          | Total time for Hartree multipole sum       :        5.042 s           5.044 s
          | Total time for total energy evaluation     :        0.002 s           0.001 s
          | Total time for scaled ZORA corrections     :        0.000 s           0.000 s

          Partial memory accounting:
          | Residual value for overall tracked memory usage across tasks:     0.000000 MB (should be 0.000000 MB)
          | Peak values for overall tracked memory usage:
          |   Minimum:       20.033 MB (on task 1 after allocating d_wave)
          |   Maximum:       20.979 MB (on task 0 after allocating d_wave)
          |   Average:       20.447 MB
          | Largest tracked array allocation:
          |   Minimum:        5.130 MB (density_matrix_con on task 1)
          |   Maximum:        5.538 MB (density_matrix_con on task 0)
          |   Average:        5.251 MB
          Note:  These values currently only include a subset of arrays which are explicitly tracked.
          The "true" memory usage will be greater.

          Have a nice day.
------------------------------------------------------------
"""

    with open("aims.out", "w") as fd:
        fd.write(output)



# creates:  nice.png

import numpy as np

from ase import Atoms
from ase.io import write
from ase.utils import hsv

atoms = Atoms('Ag', cell=(2.7, 2.7, 2.7), pbc=True) * (18, 8, 8)

# view with ASE-GUI
# view(atoms)
rotation = '-70x, -20y, -2z'  # found using ASE-GUI menu 'view -> rotate'

# Make colors
colors = hsv(atoms.positions[:, 0])

# Textures
tex = ['jmol', ] * 288 + ['glass', ] * 288 + ['ase3', ] * 288 + ['vmd', ] * 288


# Keywords that exist for eps, png, and povs
generic_projection_settings = {
    'rotation': rotation,
    'colors': colors,
    'radii': None,
}

povray_settings = {  # For povray files only
    'display': False,  # Display while rendering
    'pause': False,  # Pause when done rendering (only if display)
    'transparent': False,  # Transparent background
    'canvas_width': None,  # Width of canvas in pixels
    'canvas_height': None,  # Height of canvas in pixels
    'camera_dist': 50.,   # Distance from camera to front atom
    'image_plane': None,  # Distance from front atom to image plane
    # (focal depth for perspective)
    'camera_type': 'perspective',  # perspective, ultra_wide_angle
    'point_lights': [],             # [[loc1, color1], [loc2, color2],...]
    'area_light': [(2., 3., 40.),  # location
                   'White',       # color
                   .7, .7, 3, 3],  # width, height, Nlamps_x, Nlamps_y
    'background': 'White',        # color
    'textures': tex,  # Length of atoms list of texture names
    'celllinewidth': 0.05,  # Radius of the cylinders representing the cell
}

# Make flat png file
#write('flat.png', atoms, **kwargs)

# Make the color of the glass beads semi-transparent
colors2 = np.zeros((1152, 4))
colors2[:, :3] = colors
colors2[288: 576, 3] = 0.95
generic_projection_settings['colors'] = colors2

# Make the raytraced image
# first write the configuration files, then call the external povray executable
renderer = write('nice.pov', atoms,
                 **generic_projection_settings,
                 povray_settings=povray_settings)
renderer.render()



"""Test module for explicitly unittesting parts of the VASP calculator"""

import os
import sys
import pytest

from ase.build import molecule
from ase.calculators.calculator import CalculatorSetupError, get_calculator_class
from ase.calculators.vasp import Vasp
from ase.calculators.vasp.vasp import check_atoms, check_pbc, check_cell, check_atoms_type


@pytest.fixture
def atoms():
    return molecule('H2', vacuum=5, pbc=True)


@pytest.fixture(autouse=True)
def always_mock_calculate(mock_vasp_calculate):
    """No tests in this module may execute VASP"""
    yield


def test_verify_no_run():
    """Verify that we get an error if we try and execute the calculator,
    due to the fixture.
    """
    calc = Vasp()
    with pytest.raises(AssertionError):
        calc._run()


def test_check_atoms(atoms):
    """Test checking atoms passes for a good atoms object"""
    check_atoms(atoms)
    check_pbc(atoms)
    check_cell(atoms)


@pytest.mark.parametrize(
    'bad_atoms',
    [
        None,
        'a_string',
        # We cannot handle lists of atoms either
        [molecule('H2', vacuum=5)],
    ])
def test_not_atoms(bad_atoms):
    """Check that passing in objects which are not
    actually Atoms objects raises a setup error """

    with pytest.raises(CalculatorSetupError):
        check_atoms_type(bad_atoms)
    with pytest.raises(CalculatorSetupError):
        check_atoms(bad_atoms)

    # Test that error is also raised properly when launching
    # from calculator
    calc = Vasp()
    with pytest.raises(CalculatorSetupError):
        calc.calculate(atoms=bad_atoms)


@pytest.mark.parametrize('pbc', [
    3 * [False],
    [True, False, True],
    [False, True, False],
])
def test_bad_pbc(atoms, pbc):
    """Test handling of PBC"""
    atoms.pbc = pbc

    check_cell(atoms)  # We have a cell, so this should not raise

    # Check that our helper functions raises the expected error
    with pytest.raises(CalculatorSetupError):
        check_pbc(atoms)
    with pytest.raises(CalculatorSetupError):
        check_atoms(atoms)

    # Check we also raise in the calculator when launching
    # a calculation, but before VASP is actually executed
    calc = Vasp()
    atoms.calc = calc
    with pytest.raises(CalculatorSetupError):
        atoms.get_potential_energy()


def test_vasp_no_cell(testdir):
    """Check missing cell handling."""
    # Molecules come with no unit cell
    atoms = molecule('CH4')
    # We should not have a cell
    assert atoms.cell.rank == 0

    with pytest.raises(CalculatorSetupError):
        check_cell(atoms)
    with pytest.raises(CalculatorSetupError):
        check_atoms(atoms)

    with pytest.raises(RuntimeError):
        atoms.write('POSCAR')

    calc = Vasp()
    atoms.calc = calc
    with pytest.raises(CalculatorSetupError):
        atoms.get_total_energy()


def test_vasp_name():
    """Test the calculator class has the expected name"""
    expected = 'vasp'
    assert Vasp.name == expected  # Test class attribute
    assert Vasp().name == expected  # Ensure instance attribute hasn't changed


def test_vasp_get_calculator():
    cls_ = get_calculator_class('vasp')

    assert cls_ == Vasp

    # Test we get the correct calculator when loading from name
    assert get_calculator_class(Vasp.name) == cls_


@pytest.mark.parametrize('envvar', Vasp.env_commands)
def test_make_command_envvar(envvar, monkeypatch, clear_vasp_envvar):
    """Test making a command based on the environment variables"""
    # Environment should be cleared by the "clear_vasp_envvar" fixture
    assert envvar not in os.environ
    cmd_str = 'my command'
    monkeypatch.setenv(envvar, cmd_str)
    calc = Vasp()

    cmd = calc.make_command()
    if envvar == 'VASP_SCRIPT':
        # This envvar uses the `sys.exe` to create the command
        exe = sys.executable
        assert cmd == f'{exe} {cmd_str}'
    else:
        # We just use the exact string
        assert cmd == cmd_str


def test_make_command_no_envvar(monkeypatch, clear_vasp_envvar):
    """Test we raise when making a command with not enough information"""
    # Environment should be cleared by the "clear_vasp_envvar" fixture
    calc = Vasp()
    with pytest.raises(CalculatorSetupError):
        calc.make_command()


def test_make_command_explicit(monkeypatch):
    """Test explicitly passing a command to the calculator"""
    # The command should be whatever we put in, if we explicitly set something
    # envvars should not matter
    for envvar in Vasp.env_commands:
        # Populate the envvars with some strings, they should not matter
        monkeypatch.setenv(envvar, 'something')
    calc = Vasp()
    my_cmd = 'my command'
    cmd = calc.make_command(my_cmd)
    assert cmd == my_cmd



import pytest
import numpy as np
from ase.geometry.bravais_type_engine import generate_niggli_op_table


ref_info = {
    'FCC': 1,
    'BCC': 1,
    'CUB': 1,
    'TET': 2,
    'BCT': 5,
    'HEX': 2,
    'ORC': 1,
    'ORCC': 5,
    'ORCF': 2,
    'ORCI': 4,
    'RHL': 3,
    #'MCL': 15,
    #'MCLC': 27,
    #'TRI': 19,
}


# We disable the three lattices that have infinite reductions.
# Maybe we can test those, but not today.
assert len(ref_info) == 14 - 3


def ref_info_iter():
    for key, val in ref_info.items():
        yield key, val


@pytest.mark.parametrize('lattice_name,ref_nops', ref_info_iter())
def test_generate_niggli_table(lattice_name, ref_nops):
    length_grid = np.logspace(-1, 1, 60)
    angle_grid = np.linspace(30, 120, 90 + 59)
    table = generate_niggli_op_table(lattices=[lattice_name],
                                     angle_grid=angle_grid,
                                     length_grid=length_grid)
    for key in table:
        print('{}: {}'.format(key, len(table[key])))

    mappings = table[lattice_name]
    assert len(mappings) == ref_nops



def test_mopac():
    """Test H2 molecule atomization with MOPAC."""
    from ase.build import molecule
    from ase.calculators.mopac import MOPAC
    from ase.optimize import BFGS
    h2 = molecule('H2', calculator=MOPAC(label='h2'))
    BFGS(h2, trajectory='h2.traj').run(fmax=0.01)
    e2 = h2.get_potential_energy()
    h1 = h2.copy()
    del h1[1]
    h1.set_initial_magnetic_moments([1])
    h1.calc = MOPAC(label='h1')
    e1 = h1.get_potential_energy()
    d = h2.get_distance(0, 1)
    ea = 2 * e1 - e2
    print(d, ea)
    assert abs(d - 0.759) < 0.001
    assert abs(ea - 5.907) < 0.001
    h2o = molecule('H2O', calculator=MOPAC(label='h2o', tasks='GRADIENTS'))
    h2o.get_potential_energy()
    print('dipole:', h2o.get_dipole_moment())
    atoms = MOPAC.read_atoms('h2')
    print('magmom:', atoms.calc.get_magnetic_moment())
    print('PM7 homo lumo:', atoms.calc.get_homo_lumo_levels())
    atoms.calc.set(method='AM1')
    atoms.get_potential_energy()
    print('AM1 homo lumo:', atoms.calc.get_homo_lumo_levels())
    calc = MOPAC(restart='h1')
    print('magmom:', calc.get_magnetic_moment())



import numpy.random as random
import numpy as np
import pytest
from ase import Atoms
from ase.neighborlist import (NeighborList, PrimitiveNeighborList,
                              NewPrimitiveNeighborList)
from ase.build import bulk


def count(nl, atoms):
    c = np.zeros(len(atoms), int)
    R = atoms.get_positions()
    cell = atoms.get_cell()
    d = 0.0
    for a in range(len(atoms)):
        i, offsets = nl.get_neighbors(a)
        for j in i:
            c[j] += 1
        c[a] += len(i)
        d += (((R[i] + np.dot(offsets, cell) - R[a])**2).sum(1)**0.5).sum()
    return d, c


# scipy sparse uses matrix subclass internally
@pytest.mark.filterwarnings('ignore:the matrix subclass')
@pytest.mark.slow
def test_supercell():
    atoms = Atoms(numbers=range(10),
                  cell=[(0.2, 1.2, 1.4),
                        (1.4, 0.1, 1.6),
                        (1.3, 2.0, -0.1)])
    atoms.set_scaled_positions(3 * random.random((10, 3)) - 1)

    for sorted in [False, True]:
        for p1 in range(2):
            for p2 in range(2):
                for p3 in range(2):
                    # print(p1, p2, p3)
                    atoms.set_pbc((p1, p2, p3))
                    nl = NeighborList(atoms.numbers * 0.2 + 0.5,
                                      skin=0.0, sorted=sorted)
                    nl.update(atoms)
                    d, c = count(nl, atoms)
                    atoms2 = atoms.repeat((p1 + 1, p2 + 1, p3 + 1))
                    nl2 = NeighborList(atoms2.numbers * 0.2 + 0.5,
                                       skin=0.0, sorted=sorted)
                    nl2.update(atoms2)
                    d2, c2 = count(nl2, atoms2)
                    c2.shape = (-1, 10)
                    dd = d * (p1 + 1) * (p2 + 1) * (p3 + 1) - d2
                    assert abs(dd) < 1e-10
                    assert not (c2 - c).any()


def test_H2():
    h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
    nl = NeighborList([0.5, 0.5], skin=0.1, sorted=True, self_interaction=False)
    nl2 = NeighborList([0.5, 0.5], skin=0.1, sorted=True,
                       self_interaction=False,
                       primitive=NewPrimitiveNeighborList)
    assert nl2.update(h2)
    assert nl.update(h2)
    assert not nl.update(h2)
    assert (nl.get_neighbors(0)[0] == [1]).all()
    m = np.zeros((2, 2))
    m[0, 1] = 1
    assert np.array_equal(nl.get_connectivity_matrix(sparse=False), m)
    assert np.array_equal(nl.get_connectivity_matrix(sparse=True).todense(), m)
    assert np.array_equal(nl.get_connectivity_matrix().todense(),
                          nl2.get_connectivity_matrix().todense())

    h2[1].z += 0.09
    assert not nl.update(h2)
    assert (nl.get_neighbors(0)[0] == [1]).all()

    h2[1].z += 0.09
    assert nl.update(h2)
    assert (nl.get_neighbors(0)[0] == []).all()
    assert nl.nupdates == 2


def test_H2_shape_and_type():
    h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
    nl = NeighborList([0.1, 0.1], skin=0.1, bothways=True,
                      self_interaction=False)
    assert nl.update(h2)
    assert nl.get_neighbors(0)[1].shape == (0, 3)
    assert nl.get_neighbors(0)[1].dtype == int


def test_fcc():
    x = bulk('X', 'fcc', a=2**0.5)

    nl = NeighborList([0.5], skin=0.01, bothways=True, self_interaction=False)
    nl.update(x)
    assert len(nl.get_neighbors(0)[0]) == 12

    nl = NeighborList([0.5] * 27, skin=0.01, bothways=True,
                      self_interaction=False)
    nl.update(x * (3, 3, 3))
    for a in range(27):
        assert len(nl.get_neighbors(a)[0]) == 12
    assert not np.any(nl.get_neighbors(13)[1])

    c = 0.0058
    for NeighborListClass in [PrimitiveNeighborList, NewPrimitiveNeighborList]:
        nl = NeighborListClass([c, c],
                               skin=0.0,
                               sorted=True,
                               self_interaction=False,
                               use_scaled_positions=True)
        nl.update([True, True, True],
                  np.eye(3) * 7.56,
                  np.array([[0, 0, 0],
                            [0, 0, 0.99875]]))
        n0, d0 = nl.get_neighbors(0)
        n1, d1 = nl.get_neighbors(1)
        # != is xor
        assert (np.all(n0 == [0]) and np.all(d0 == [0, 0, 1])) != \
            (np.all(n1 == [1]) and np.all(d1 == [0, 0, -1]))


def test_empty_neighbor_list():
    # Test empty neighbor list
    nl = PrimitiveNeighborList([])
    nl.update([True, True, True],
              np.eye(3) * 7.56,
              np.zeros((0, 3)))


def test_hexagonal_cell_and_large_cutoff():
    # Test hexagonal cell and large cutoff
    pbc_c = np.array([True, True, True])
    cutoff_a = np.array([8.0, 8.0])
    cell_cv = np.array([[0., 3.37316113, 3.37316113],
                        [3.37316113, 0., 3.37316113],
                        [3.37316113, 3.37316113, 0.]])
    spos_ac = np.array([[0., 0., 0.],
                        [0.25, 0.25, 0.25]])

    nl = PrimitiveNeighborList(cutoff_a, skin=0.0, sorted=True,
                               use_scaled_positions=True)
    nl2 = NewPrimitiveNeighborList(cutoff_a, skin=0.0, sorted=True,
                                   use_scaled_positions=True)
    nl.update(pbc_c, cell_cv, spos_ac)
    nl2.update(pbc_c, cell_cv, spos_ac)

    a0, offsets0 = nl.get_neighbors(0)
    b0 = np.zeros_like(a0)
    d0 = np.dot(spos_ac[a0] + offsets0 - spos_ac[0], cell_cv)
    a1, offsets1 = nl.get_neighbors(1)
    d1 = np.dot(spos_ac[a1] + offsets1 - spos_ac[1], cell_cv)
    b1 = np.ones_like(a1)

    a = np.concatenate([a0, a1])
    b = np.concatenate([b0, b1])
    d = np.concatenate([d0, d1])
    _a = np.concatenate([a, b])
    _b = np.concatenate([b, a])
    a = _a
    b = _b
    d = np.concatenate([d, -d])

    a0, offsets0 = nl2.get_neighbors(0)
    d0 = np.dot(spos_ac[a0] + offsets0 - spos_ac[0], cell_cv)
    b0 = np.zeros_like(a0)
    a1, offsets1 = nl2.get_neighbors(1)
    d1 = np.dot(spos_ac[a1] + offsets1 - spos_ac[1], cell_cv)
    b1 = np.ones_like(a1)

    a2 = np.concatenate([a0, a1])
    b2 = np.concatenate([b0, b1])
    d2 = np.concatenate([d0, d1])
    _a2 = np.concatenate([a2, b2])
    _b2 = np.concatenate([b2, a2])
    a2 = _a2
    b2 = _b2
    d2 = np.concatenate([d2, -d2])

    i = np.argsort(d[:, 0] + d[:, 1] * 1e2 + d[:, 2] * 1e4 + a * 1e6)
    i2 = np.argsort(d2[:, 0] + d2[:, 1] * 1e2 + d2[:, 2] * 1e4 + a2 * 1e6)

    assert np.all(a[i] == a2[i2])
    assert np.all(b[i] == b2[i2])
    assert np.allclose(d[i], d2[i2])


def test_small_cell_and_large_cutoff():
    # See: https://gitlab.com/ase/ase/-/issues/441
    cutoff = 50

    atoms = bulk('Cu', 'fcc', a=3.6)
    atoms *= (2, 2, 2)
    atoms.set_pbc(False)
    radii = cutoff * np.ones(len(atoms.get_atomic_numbers()))

    neighborhood_new = NeighborList(
        radii, skin=0.0, self_interaction=False, bothways=True,
        primitive=NewPrimitiveNeighborList
    )
    neighborhood_old = NeighborList(
        radii, skin=0.0, self_interaction=False, bothways=True,
        primitive=PrimitiveNeighborList
    )

    neighborhood_new.update(atoms)
    neighborhood_old.update(atoms)

    n0, d0 = neighborhood_new.get_neighbors(0)
    n1, d1 = neighborhood_old.get_neighbors(0)

    assert np.all(n0 == n1)
    assert np.all(d0 == d1)



import pytest
from ase.utils.filecache import MultiFileJSONCache, CombinedJSONCache, Locked


pytestmark = pytest.mark.usefixtures('testdir')


@pytest.fixture
def cache():
    return MultiFileJSONCache('cache')


def sample_dict():
    return {'hello': [1, 2, 3], 'world': 'grumble'}


def test_basic(cache):
    assert len(cache) == 0

    cache['hello'] = 'grumble'
    assert len(cache) == 1
    assert 'hello' in cache

    grumble = cache.pop('hello')
    assert grumble == 'grumble'
    assert 'hello' not in cache
    assert len(cache) == 0


@pytest.mark.parametrize('dct', [
    {},
    sample_dict(),
])
def test_cache(dct, cache):
    cache.update(dct)
    assert dict(cache) == dct


def test_combine(cache):
    dct = sample_dict()
    cache.update(dct)
    combined = cache.combine()
    assert dict(combined) == dct


def test_split():
    dct = sample_dict()
    combined = CombinedJSONCache.dump_cache('cache', dct)
    assert dict(combined) == dct
    cache = combined.split()
    assert dict(cache) == dct
    assert len(combined) == 0


def test_lock(cache):
    with cache.lock('hello'):
        # When element is locked but nothing is written, the
        # cache is defined to "contain" None
        assert 'hello' in cache
        assert cache['hello'] is None

        # Other keys should function as normal:
        cache['xx'] = 1
        assert cache['xx'] == 1


def test_already_locked(cache):
    with cache.lock('hello') as handle:
        assert handle is not None
        with cache.lock('hello') as otherhandle:
            assert otherhandle is None

        with pytest.raises(Locked):
            cache['hello'] = 'world'


def test_no_overwrite_combine(cache):
    cache.combine()
    with pytest.raises(RuntimeError, match='Already exists'):
        cache.combine()



from math import sqrt
from ase import Atoms, Atom
from ase.constraints import FixAtoms
from ase.calculators.emt import EMT
from ase.optimize import QuasiNewton
from ase.io import read
from ase.visualize import view


def test_replay(testdir):
    # Distance between Cu atoms on a (100) surface:
    d = 3.6 / sqrt(2)
    a = Atoms('Cu',
              positions=[(0, 0, 0)],
              cell=(d, d, 1.0),
              pbc=(True, True, False))
    a *= (2, 2, 1)  # 2x2 (100) surface-cell

    # Approximate height of Ag atom on Cu(100) surfece:
    h0 = 2.0
    a += Atom('Ag', (d / 2, d / 2, h0))

    if 0:
        view(a)

    constraint = FixAtoms(range(len(a) - 1))
    a.calc = EMT()
    a.set_constraint(constraint)

    with QuasiNewton(a, trajectory='AgCu1.traj', logfile='AgCu1.log') as dyn1:
        dyn1.run(fmax=0.1)

    a = read('AgCu1.traj')
    a.calc = EMT()
    print(a.constraints)

    with QuasiNewton(a, trajectory='AgCu2.traj', logfile='AgCu2.log') as dyn2:
        dyn2.replay_trajectory('AgCu1.traj')
        dyn2.run(fmax=0.01)



from ase.visualize import view
from ase.io.opls import OPLSStructure

s = OPLSStructure('172_mod.xyz')  # 172_mod.xyz if the file name for the structure above
view(s)  # view with real elements
elements = {'CT': 'Si', 'HC': 'H', 'H1': 'He'}
view(s.colored(elements))  # view with fake elements



def test_gaussian_cmdline(cli):
    from ase.db import connect
    from ase.io import read
    from ase.io.jsonio import read_json

    cli.shell("""\
    ase build O O.xyz && ase run gaussian O.xyz -o gaussian_cmdline.json &&
    ase build O2 O2.xyz && ase run gaussian O2.xyz -o gaussian_cmdline.json""",
              'gaussian')
    c = connect('gaussian_cmdline.json')
    dct = read_json('gaussian_cmdline.json')
    for index, name in enumerate(['O', 'O2']):
        d = c.get(index + 1)
        id = d.id
        e1 = d.energy
        e2 = c.get_atoms(id).get_potential_energy()
        e3 = read(name + '.log').get_potential_energy()
        e4 = dct[id]['energy']
        assert e1 == e2 == e3 == e4
        print(e1)
    ae = 2 * c.get(1).energy - c.get(2).energy
    assert abs(ae - 0.65376) < 1e-3



from ase import Atoms
from ase.calculators.emt import EMT
from ase.md import VelocityVerlet
from ase.io import Trajectory


def test_md(testdir):
    a = 3.6
    b = a / 2
    fcc = Atoms('Cu', positions=[(0, 0, 0)],
                cell=[(0, b, b), (b, 0, b), (b, b, 0)],
                pbc=1)
    fcc *= (2, 1, 1)
    fcc.calc = EMT()
    fcc.set_momenta([(0.9, 0.0, 0.0), (-0.9, 0, 0)])

    def f():
        print(fcc.get_potential_energy(), fcc.get_total_energy())

    with VelocityVerlet(fcc, timestep=0.1) as md:
        md.attach(f)
        with Trajectory('Cu2.traj', 'w', fcc) as traj:
            md.attach(traj.write, interval=3)
            md.run(steps=20)

    with Trajectory('Cu2.traj', 'r') as traj:
        traj[-1]

    # Really?? No assertion at all?



import pytest
from ase.build import bulk
from ase.test.factories import ObsoleteFactoryWrapper


omx_par = {'definition_of_atomic_species': [['Al', 'Al8.0-p1', 'Al_CA13'],
                                            ['O', 'O6.0-p1', 'O_CA13']]}


required = {'aims': dict(sc_accuracy_rho=5.e-3)}


calc = pytest.mark.calculator


@calc('abinit', ecut=200, toldfe=0.0001, chksymbreak=0)
@calc('openmx', **omx_par)
@calc('elk', rgkmax=5.0)
def test_al(factory):
    run(factory)


@pytest.mark.parametrize('name', sorted(required))
def test_al_old(name):
    factory = ObsoleteFactoryWrapper(name)
    run(factory)


def run(factory):
    name = factory.name
    par = required.get(name, {})
    # What on earth does kpts=1.0 mean?  Was failing, I changed it.  --askhl
    # Disabled GPAW since it was failing anyway. --askhl
    kpts = [2, 2, 2]
    calc = factory.calc(label=name, xc='LDA', kpts=kpts, **par)
    al = bulk('AlO', crystalstructure='rocksalt', a=4.5)
    al.calc = calc
    e = al.get_potential_energy()
    calc.set(xc='PBE', kpts=kpts)
    epbe = al.get_potential_energy()
    print(e, epbe)
    calc = factory.calc(restart=name)
    print(calc.parameters, calc.results, calc.atoms)
    assert not calc.calculation_required(al, ['energy'])
    al = calc.get_atoms()
    print(al.get_potential_energy())
    label = 'dir/' + name + '-2'
    calc = factory.calc(label=label, atoms=al, xc='LDA', kpts=kpts,
                        **par)
    print(al.get_potential_energy())



from ase.units import fs
from ase.build import bulk
from ase.md import Langevin
from ase.md.fix import FixRotation
from ase.utils import seterr
from ase.md.velocitydistribution import (MaxwellBoltzmannDistribution,
                                         Stationary)
import numpy as np


def check_inertia(atoms):
    m, v = atoms.get_moments_of_inertia(vectors=True)
    print("Moments of inertia:")
    print(m)
    # There should be one vector in the z-direction
    n = 0
    delta = 1e-2
    for a in v:
        if (abs(a[0]) < delta
            and abs(a[1]) < delta
            and abs(abs(a[2]) - 1.0) < delta):

            print("Vector along z:", a)
            n += 1
        else:
            print("Vector not along z:", a)
    assert n == 1


def test_fixrotation_asap(asap3):
    rng = np.random.RandomState(123)

    with seterr(all='raise'):
        atoms = bulk('Au', cubic=True).repeat((3, 3, 10))
        atoms.pbc = False
        atoms.center(vacuum=5.0 + np.max(atoms.cell) / 2)
        print(atoms)
        atoms.calc = asap3.EMT()
        MaxwellBoltzmannDistribution(atoms, temperature_K=300, force_temp=True,
                                     rng=rng)
        Stationary(atoms)
        check_inertia(atoms)
        with Langevin(
                atoms,
                timestep=20 * fs,
                temperature_K=300,
                friction=1e-3,
                logfile='-',
                loginterval=500,
                rng=rng
        ) as md:
            fx = FixRotation(atoms)
            md.attach(fx)
            md.run(steps=1000)
        check_inertia(atoms)
